"""
    html -> creole Emitter
    ~~~~~~~~~~~~~~~~~~~~~~


    :copyleft: 2008-2011 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import posixpath

from creole.shared.base_emitter import BaseEmitter


class CreoleEmitter(BaseEmitter):
    """
    Build from a document_tree (html2creole.parser.HtmlParser instance) a
    creole markup text.
    """

    def __init__(self, document_tree, strict=False, *args, **kwargs):
        self.strict = strict
        super().__init__(document_tree, *args, **kwargs)

        self.table_head_prefix = "= "
        self.table_auto_width = True

    def emit(self):
        """Emit the document represented by self.root DOM tree."""
        return self.emit_node(self.root).strip()  # FIXME

    # --------------------------------------------------------------------------

    def blockdata_pre_emit(self, node):
        """ pre block -> with newline at the end """
        return "{{{%s}}}\n" % self.deentity.replace_all(node.content)

    def inlinedata_pre_emit(self, node):
        """ a pre inline block -> no newline at the end """
        return "{{{%s}}}" % self.deentity.replace_all(node.content)

    def blockdata_pass_emit(self, node):
        return f"{node.content}\n\n"

    # --------------------------------------------------------------------------

    def p_emit(self, node):
        result = self.emit_children(node)
        if self._inner_list == "":
            result += "\n\n"
        return result

    def br_emit(self, node):
        if self._inner_list != "":
            return "\\\\"
        else:
            return "\n"

    def headline_emit(self, node):
        return f"{'=' * node.level} {self.emit_children(node)}\n\n"

    # --------------------------------------------------------------------------

    def strong_emit(self, node):
        # Atlassian uses a single * for bold text
        return self._typeface(node, key="*")
    b_emit = strong_emit
    big_emit = strong_emit

    def i_emit(self, node):
        return self._typeface(node, key="//")
    em_emit = i_emit

    def tt_emit(self, node):
        return self._typeface(node, key="##")

    def sup_emit(self, node):
        return self._typeface(node, key="^^")

    def sub_emit(self, node):
        return self._typeface(node, key=",,")

    def u_emit(self, node):
        return self._typeface(node, key="__")

    def small_emit(self, node):
        return self._typeface(node, key="--")

    def del_emit(self, node):
        return self._typeface(node, key="~~")
    strike_emit = del_emit

    # --------------------------------------------------------------------------

    def hr_emit(self, node):
        return "----\n\n"

    def a_emit(self, node):
        link_text = self.emit_children(node)
        try:
            url = node.attrs["href"]
        except KeyError:
            # e.g.: <a name="anchor-one">foo</a>
            return link_text
        if link_text == url:
            return f"[[{url}]]"
        else:
            # Atlassian style links
            # [link title|http://example.com] 
            return f"[{link_text}|{url}]"

    def img_emit(self, node):
        src = node.attrs["src"]

        if src.split(':')[0] == 'data':
            return ""

        title = node.attrs.get("title", "")
        alt = node.attrs.get("alt", "")
        width = node.attrs.get("height", None)
        height = node.attrs.get("width", None)
        if len(alt) > len(title):  # Use the longest one
            text = alt
        else:
            text = title

        if text == "":  # Use filename as picture text
            text = posixpath.basename(src)

        if not self.strict:
            if width and height:
                return f"{{{{{src}|{text}|{width}x{height}}}}}"

        return f"{{{{{src}|{text}}}}}"

    # --------------------------------------------------------------------------

    def ul_emit(self, node):
        return self._list_emit(node, list_type="*")

    def ol_emit(self, node):
        return self._list_emit(node, list_type="#")

    # --------------------------------------------------------------------------

    def div_emit(self, node):
        return self._emit_content(node)

    def span_emit(self, node):
        return self._emit_content(node)



"""
    Creole unittest macros
    ~~~~~~~~~~~~~~~~~~~~~~

    Note: all mecro functions must return unicode!

    :copyleft: 2008-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import json


def unittest_macro1(**kwargs):
    """
    >>> unittest_macro1(foo="bar")
    '[test macro1 - kwargs: foo="bar"]'

    >>> unittest_macro1()
    '[test macro1 - kwargs: ]'

    >>> unittest_macro1(a=1,b=2)
    '[test macro1 - kwargs: a=1,b=2]'
    """
    kwargs = ','.join('{}={}'.format(k, json.dumps(v)) for k, v in sorted(kwargs.items()))
    return f"[test macro1 - kwargs: {kwargs}]"


def unittest_macro2(char, text):
    """
    >>> unittest_macro2(char="|", text="a\\nb")
    'a|b'
    """
    return char.join(text.split())



"""
    unittest subprocess helper
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2015-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import os
import shutil
import subprocess
import sys
from pathlib import Path


class SubprocessMixin:

    def subprocess(self, popen_args):
        assert isinstance(popen_args, (tuple, list))

        print("Call:", popen_args)

        # Expand PATH
        bin_path = str(Path(sys.executable).parent)
        if bin_path not in os.environ["PATH"]:
            # .../venv/bin will be not in PATH in tests, just add it
            # so that installed [tool.poetry.scripts] will be found
            os.environ["PATH"] += os.pathsep + bin_path

        # Check if executeable will be found
        prog = popen_args[0]
        cmd = shutil.which(prog)
        assert cmd is not None, f'{prog!r} not found in PATH: {os.environ.get("PATH")!r}!'

        try:
            process = subprocess.Popen(
                popen_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
        except Exception as err:
            self.fail(f"Error subprocess call with {popen_args!r}: {err}")

        stdout, stderr = process.communicate()
        retcode = process.poll()

        print(f"return code: {retcode!r}")
        print(f"stdout: {stdout!r}")
        print(f"stderr: {stderr!r}")

        stdout = stdout.strip()
        print("=" * 100)
        print("stdout:")
        print("-" * 100)
        print(stdout)
        print("-" * 100)
        return popen_args, retcode, stdout

    def assertSubprocess(self, popen_args, retcode, stdout):
        popen_args2, retcode2, stdout2 = self.subprocess(popen_args)
        stdout = stdout.strip()
        try:
            self.assertEqual(stdout, stdout2, "stdout wrong:")
            self.assertEqual(retcode, retcode2, "return code wrong:")
        except AssertionError as err:
            msg = (
                "Error: %s"
                "call via subprocess: %s\n"
                "return code........: %r\n"
                " ---------- [stdout] ---------- \n"
                "%s\n"
                "-------------------------------"
            ) % (
                err,
                repr(popen_args2), retcode2,
                stdout2,
            )
            self.fail(msg)



"""
    python-creole exceptions
    ~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2011 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


class DocutilsImportError(ImportError):
    pass



"""
    html2creole tests
    ~~~~~~~~~~~~~~~~~

    special html to creole convert tests, witch can't tests in "cross compare"


    :copyleft: 2008-2011 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import unittest

from creole import html2creole
from creole.shared.unknown_tags import (
    escape_unknown_nodes,
    raise_unknown_node,
    transparent_unknown_nodes,
    use_html_macro,
)
from creole.tests.utils.base_unittest import BaseCreoleTest


class TestHtml2Creole(unittest.TestCase):
    """
    Tests around html2creole API.
    """
    pass


class TestHtml2CreoleMarkup(BaseCreoleTest):

    #    def assertCreole(self, raw_markup, raw_html, debug=False, **kwargs):
    #        self.assert_html2creole(raw_markup, raw_html, debug=debug, **kwargs)

    # --------------------------------------------------------------------------

    def test_not_used(self):
        """
        Some other html tags -> convert.
        """
        self.assert_html2creole(r"""
            **Bold text**
            **Big text**
            //em tag//
            //italic//
        """, """
            <p><b>Bold text</b><br />
            <big>Big text</big><br />
            <em>em tag</em><br />
            <i>italic</i></p>
        """)

    # --------------------------------------------------------------------------

    def test_raise_unknown_node(self):
        """
        Test creole.html2creole.raise_unknown_node callable:
        Raise NotImplementedError on unknown tags.
        """
        self.assertRaises(NotImplementedError,
                          html2creole,
                          html_string="<unknwon>",
                          unknown_emit=raise_unknown_node
                          )

    def test_use_html_macro(self):
        """
        Test creole.html2creole.use_html_macro callable:
        Use the <<html>> macro to mask unknown tags.
        """
        self.assert_html2creole(r"""
            111 <<html>><unknown><</html>>foo<<html>></unknown><</html>> 222
            333<<html>><unknown foo1="bar1" foo2="bar2"><</html>>foobar<<html>></unknown><</html>>444

            555<<html>><unknown /><</html>>666
        """, """
            <p>111 <unknown>foo</unknown> 222<br />
            333<unknown foo1="bar1" foo2="bar2">foobar</unknown>444</p>

            <p>555<unknown />666</p>
        """,
                                unknown_emit=use_html_macro
                                )

    def test_escape_unknown_nodes(self):
        """
        Test creole.html2creole.escape_unknown_nodes callable:
        All unknown tags should be escaped.
        """
        self.assert_html2creole(r"""
            111 &lt;unknown&gt;foo&lt;/unknown&gt; 222
            333&lt;unknown foo1="bar1" foo2="bar2"&gt;foobar&lt;/unknown&gt;444

            555&lt;unknown /&gt;666
        """, """
            <p>111 <unknown>foo</unknown> 222<br />
            333<unknown foo1="bar1" foo2="bar2">foobar</unknown>444</p>

            <p>555<unknown />666</p>
        """,
                                unknown_emit=escape_unknown_nodes
                                )

    def test_escape_unknown_nodes2(self):
        """
        HTMLParser has problems with <script> tags.
        See: http://bugs.python.org/issue670664
        """
        self.assert_html2creole(r"""
            &lt;script&gt;var js_sha_link='<p>***</p>';&lt;/script&gt;
        """, """
            <script>
            var js_sha_link='<p>***</p>';
            </script>
        """,
                                unknown_emit=escape_unknown_nodes
                                )

    def test_transparent_unknown_nodes(self):
        """
        Test creole.html2creole.transparent_unknown_nodes callable:
        All unknown tags should be "transparent" and show only
        their child nodes' content.
        """
        self.assert_html2creole(r"""
            //baz//, **quux**
        """, """
            <form class="foo" id="bar"><label><em>baz</em></label>, <strong>quux</strong></form>
        """, unknown_emit=transparent_unknown_nodes
                                )

    def test_transparent_unknown_nodes2(self):
        """
        HTMLParser has problems with <script> tags.
        See: http://bugs.python.org/issue670664
        """
        self.assert_html2creole(r"""
            FOO var a='<em>STRONG</em>'; BAR
        """, """
            <p>FOO <script>var a='<em>STRONG</em>';</script> BAR</p>
        """, unknown_emit=transparent_unknown_nodes
                                )

    def test_transparent_unknown_nodes_block_elements(self):
        """
        Test that block elements insert linefeeds into the stream.
        """
        self.assert_html2creole(r"""
            //baz//,

            **quux**

            spam, ham, and eggs
        """, """
            <div><em>baz</em>,</div> <fieldset><strong>quux</strong></fieldset>
            <span>spam, </span><label>ham, </label>and eggs
        """, unknown_emit=transparent_unknown_nodes
                                )

    # --------------------------------------------------------------------------

    def test_entities(self):
        """
        Test html entities.

        copyright sign is in Latin-1 Supplement:
            http://pylucid.org/_command/144/DecodeUnicode/display/1/
        Box Drawing:
            http://pylucid.org/_command/144/DecodeUnicode/display/66/
        """
        self.assert_html2creole("""
            * less-than sign: < < <
            * greater-than sign: > > >
            * copyright sign: © ©
            * box drawing: ╬ ╬
            * german umlauts: ä ö ü
        """, """
            <ul>
            <li>less-than sign: &lt; &#60; &#x3C;</li>
            <li>greater-than sign: &gt; &#62; &#x3E;</li>
            <li>copyright sign: &#169; &#xA9;</li>
            <li>box drawing: &#9580; &#x256C;</li>
            <li>german umlauts: &auml; &ouml; &uuml;</li>
            </ul>
        """)

    def test_html_entity_nbsp(self):
        """ Non breaking spaces is not in htmlentitydefs """
        self.assert_html2creole(r"""
            a non braking space: [ ] !
        """, """
            <p>a non braking space: [&nbsp;] !</p>
        """)

    def test_html_entity_in_pre(self):
        self.assert_html2creole(r"""
            {{{<code>{% lucidTag RSS url="http url" %}</code>}}}
        """, """
            <pre><code>&#x7B;% lucidTag RSS url="http url" %&#x7D;</code></pre>
        """)

    def test_unknown_entity(self):
        """
        Test a unknown html entity.
        FIXME: What sould happend?
        """
        self.assert_html2creole(r"""
            copy&paste
        """, """
            <p>copy&paste</p>
        """)
        self.assert_html2creole(r"""
            [[/url/|Search & Destroy]]
        """, """
            <a href="/url/">Search & Destroy</a>
        """)

    def test_tbody_table(self):
        self.assert_html2creole(r"""
            Ignore 'tbody' tag in tables:

            |= Headline 1 |= Headline 2 |
            | cell one    | cell two    |
            end
        """, """
            <p>Ignore 'tbody' tag in tables:</p>
            <table>
            <tbody>
            <tr>
                <th>Headline 1</th>
                <th>Headline 2</th>
            </tr>
            <tr>
                <td>cell one</td>
                <td>cell two</td>
            </tr>
            </tbody>
            </table>
            <p>end</p>
        """)

    def test_p_table(self):
        """ strip <p> tags in table cells """
        self.assert_html2creole(r"""
            | cell one | cell two\\new line |
        """, """
            <table>
            <tr>
                <td><p>cell one</p></td>
                <td><p>cell two</p><p>new line</p><p></p></td>
            </tr>
            </table>
        """)

    def test_image(self):
        """ test image tag with different alt/title attribute """
        self.assert_html2creole(r"""
            {{foobar1.jpg|foobar1.jpg}}
            {{/foobar2.jpg|foobar2.jpg}}
            {{/path1/path2/foobar3.jpg|foobar3.jpg}}
            {{/foobar4.jpg|It's foobar 4}}
            {{/foobar5.jpg|It's foobar 5}}
            {{/foobar6.jpg|a long picture title}}
        """, """
            <p><img src="foobar1.jpg" /><br />
            <img src="/foobar2.jpg" /><br />
            <img src="/path1/path2/foobar3.jpg" /><br />
            <img src="/foobar4.jpg" alt="It's foobar 4" /><br />
            <img src="/foobar5.jpg" title="It's foobar 5" /><br />
            <img src="/foobar6.jpg" alt="short name" title="a long picture title" /><br />
            <img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA
            AAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO
            9TXL0Y4OHwAAAABJRU5ErkJggg==" alt="data uri should be disallowed" /></p>
        """)

    def test_image_with_size(self):
        """ test image tag with sizes """
        self.assert_html2creole(r"""
            {{foobar1.jpg|foobar1.jpg}}
            {{foobar2.jpg|foobar2.jpg|90x160}}
            {{foobar3.jpg|foobar3.jpg}}
        """, """
            <p><img src="foobar1.jpg" /><br />
            <img src="foobar2.jpg" width="160" height="90" /><br />
            <img src="foobar3.jpg" width="160" /></p>
        """)

    def test_image_with_size_strict(self):
        """ test image tag with sizes """
        self.assert_html2creole(r"""
            {{foobar1.jpg|foobar1.jpg}}
            {{foobar2.jpg|foobar2.jpg}}
            {{foobar3.jpg|foobar3.jpg}}
        """, """
            <p><img src="foobar1.jpg" /><br />
            <img src="foobar2.jpg" width="160" height="90" /><br />
            <img src="foobar3.jpg" width="160" /></p>
        """, strict=True)

    def test_non_closed_br(self):
        self.assert_html2creole(r"""
            one
            two
        """, """
            <p>one<br>
            two</p>
        """)

    def test_explicit_closed_br(self):
        self.assert_html2creole(r"""
            one
            two
        """, """
            <p>one<br></br>
            two</p>
        """)

    def test_newline_before_list(self):
        """
        http://code.google.com/p/python-creole/issues/detail?id=16
        """
        self.assert_html2creole(r"""
            **foo**

            * one
        """, """
            <b>foo</b><ul><li>one</li></ul>
        """)

    def test_empty_tags_are_not_escaped(self):
        self.assert_html2creole(r"""
            //baz//, **quux**
        """, """
            <div class="foo" id="bar"><span><em>baz</em></span>, <strong>quux</strong></div>
        """)

    def test_nested_listsitems_with_paragraph(self):
        self.assert_html2creole("""
            * item 1
            ** subitem 1.1
            *** subsubitem 1.1.1
            *** subsubitem 1.1.2
            ** subitem 1.2
            * item 2
            ** subitem 2.1
            """, """
            <ul>
                <li><p>item 1</p>
                    <ul>
                        <li><p>subitem 1.1</p>
                            <ul>
                                <li>subsubitem 1.1.1</li>
                                <li>subsubitem 1.1.2</li>
                            </ul>
                        </li>
                            <li><p>subitem 1.2</p></li>
                    </ul>
                </li>
                <li><p>item 2</p>
                    <ul>
                        <li>subitem 2.1</li>
                    </ul>
                </li>
            </ul>
        """)

    def test_class_in_list(self):
        """https://code.google.com/p/python-creole/issues/detail?id=19#c4"""
        self.assert_html2creole(r"""
            # foo
        """, """
            <ol class=gbtc><li>foo</li></ol>
        """)  # , debug=True)

    def test_ignore_links_without_href(self):
        """https://code.google.com/p/python-creole/issues/detail?id=19#c4"""
        self.assert_html2creole(r"""
            bar
        """, """
            <a class="foo">bar</a>
        """)  # , debug=True)

    def test_newlines_after_headlines(self):
        self.assert_html2creole(r"""
            = Headline news

            [[http://google.com|The googlezor]] is a big bad mother.
        """, """
            <h1>Headline news</h1>

            <p><a href="http://google.com">The googlezor</a> is a big bad mother.</p>
        """)

    def test_links(self):
        self.assert_html2creole(r"""
            test link: '[[internal links|link A]]' 1 and
            test link: '[[http://domain.tld|link B]]' 2.
        """, """
            <p>test link: '<a href="internal links">link A</a>' 1 and<br />
            test link: '<a href="http://domain.tld">link B</a>' 2.</p>
        """)

    def test_horizontal_rule(self):
        self.assert_html2creole(r"""
            one

            ----

            two
        """, """
            <p>one</p>
            <hr />
            <p>two</p>
        """)

    def test_nested_empty_tags(self):
        self.assert_html2creole2("TEST", "<p>TEST</p>")
        self.assert_html2creole2("TEST", "<bar><p>TEST</p></bar>")
        self.assert_html2creole2("TEST", "<foo><bar><p>TEST</p></bar></foo>")


#    def test_nowiki1(self):
#        self.assert_html2creole(r"""
#            this:
#            {{{
#            //This// does **not** get [[formatted]]
#            }}}
#            and this: {{{** <i>this</i> ** }}} not, too.
#
#            === Closing braces in nowiki:
#            {{{
#            if (x != NULL) {
#              for (i = 0; i < size; i++) {
#                if (x[i] > 0) {
#                  x[i]--;
#              }}}
#            }}}
#        """, """
#            <p>this:</p>
#            <pre>
#            //This// does **not** get [[formatted]]
#            </pre>
#            <p>and this: <tt>** &lt;i&gt;this&lt;/i&gt; ** </tt> not, too.</p>
#
#            <h3>Closing braces in nowiki:</h3>
#            <pre>
#            if (x != NULL) {
#              for (i = 0; i &lt; size; i++) {
#                if (x[i] &gt; 0) {
#                  x[i]--;
#              }}}
#            </pre>
#        """)
#
#    def test_list1(self):
#        """
#        FIXME: Two newlines between a list and the next paragraph :(
#        """
#        self.assert_html2creole(r"""
#            ==== List a:
#            * a1 item
#            ** a1.1 Force\\linebreak
#            ** a1.2 item
#            *** a1.2.1 item
#            *** a1.2.2 item
#            * a2 item
#
#
#            list 'a' end
#
#            ==== List b:
#            # b1 item
#            ## b1.2 item
#            ### b1.2.1 item
#            ### b1.2.2 Force\\linebreak1\\linebreak2
#            ## b1.3 item
#            # b2 item
#
#
#            list 'b' end
#        """, """
#            <h4>List a:</h4>
#            <ul>
#            <li>a1 item</li>
#            <ul>
#                <li>a1.1 Force
#                linebreak</li>
#                <li>a1.2 item</li>
#                <ul>
#                    <li>a1.2.1 item</li>
#                    <li>a1.2.2 item</li>
#                </ul>
#            </ul>
#            <li>a2 item</li>
#            </ul>
#            <p>list 'a' end</p>
#
#            <h4>List b:</h4>
#            <ol>
#            <li>b1 item</li>
#            <ol>
#                <li>b1.2 item</li>
#                <ol>
#                    <li>b1.2.1 item</li>
#                    <li>b1.2.2 Force
#                    linebreak1
#                    linebreak2</li>
#                </ol>
#                <li>b1.3 item</li>
#            </ol>
#            <li>b2 item</li>
#            </ol>
#            <p>list 'b' end</p>
#        """,
# debug=True
#        )
#
#    def test_list2(self):
#        """ Bold, Italics, Links, Pre in Lists """
#        self.assert_html2creole(r"""
#            * **bold** item
#            * //italic// item
#
#            # item about a [[domain.tld|page link]]
#            # {{{ //this// is **not** [[processed]] }}}
#        """, """
#            <ul>
#                <li><strong>bold</strong> item</li>
#                <li><i>italic</i> item</li>
#            </ul>
#            <ol>
#                <li>item about a <a href="domain.tld">page link</a></li>
#                <li><tt>//this// is **not** [[processed]]</tt></li>
#            </ol>
#        """,
# debug=True
#        )


if __name__ == '__main__':
    unittest.main(
        #        defaultTest="TestHtml2CreoleMarkup.test_nested_listsitems_with_paragraph"
    )



"""
    cross compare reStructuredText unittest
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Compare all similarities between:
        * rest2html (used docutils)
        * html2rest

    :copyleft: 2011-2012 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import unittest

from creole.tests.utils.base_unittest import BaseCreoleTest


class CrossCompareReStTests(BaseCreoleTest):
    def test_entities(self):
        self.cross_compare_rest(
            rest_string="""
                less-than sign: <

                greater-than sign: >
            """,
            html_string="""
                <p>less-than sign: &lt;</p>
                <p>greater-than sign: &gt;</p>
            """,
            # debug=True
        )

    def test_bullet_lists_basic(self):
        self.cross_compare_rest(
            rest_string="""
                * item 1

                * item 2

                * item 3
            """,
            html_string="""
                <ul>
                <li>item 1</li>
                <li>item 2</li>
                <li>item 3</li>
                </ul>
            """,
            # debug=True
        )

    def test_numbered_lists(self):
        self.cross_compare_rest(
            rest_string="""
            #. item 1

            #. item 2

                #. item 2.1

                #. a `link in </url/>`_ list item 2.2

            #. item 3
            """,
            html_string="""
            <ol>
            <li><p>item 1</p>
            </li>
            <li><p>item 2</p>
            <ol>
            <li>item 2.1</li>
            <li>a <a href="/url/">link in</a> list item 2.2</li>
            </ol>
            </li>
            <li><p>item 3</p>
            </li>
            </ol>
            """,
            # debug=True
        )

    def test_bullet_lists_nested(self):
        self.cross_compare_rest(
            rest_string="""
                A nested bullet lists:

                * item 1

                    * A **bold subitem 1.1** here.

                        * subsubitem 1.1.1

                        * subsubitem 1.1.2 with inline |substitution text| image.

                    * subitem 1.2

                * item 2

                    * subitem 2.1

                    * *bold 2.2*

                .. |substitution text| image:: /url/to/image.png

                Text under list.
            """,
            html_string="""
                <p>A nested bullet lists:</p>
                <ul>
                <li><p>item 1</p>
                <ul>
                <li><p>A <strong>bold subitem 1.1</strong> here.</p>
                <ul>
                <li>subsubitem 1.1.1</li>
                <li>subsubitem 1.1.2 with inline <img alt="substitution text" src="/url/to/image.png" /> image.</li>
                </ul>
                </li>
                <li><p>subitem 1.2</p>
                </li>
                </ul>
                </li>
                <li><p>item 2</p>
                <ul>
                <li>subitem 2.1</li>
                <li><em>bold 2.2</em></li>
                </ul>
                </li>
                </ul>
                <p>Text under list.</p>
            """,
            # debug=True
        )

    def test_typeface_basic(self):
        """
        http://docutils.sourceforge.net/docs/user/rst/quickref.html#inline-markup
        """
        self.cross_compare_rest(
            rest_string="""
                *emphasis* **strong**
            """,
            html_string="""
                <p><em>emphasis</em> <strong>strong</strong></p>
            """
        )

    def test_substitution_image_with_alt(self):
        self.cross_compare_rest(
            rest_string="""
                A inline |substitution text| image.

                .. |substitution text| image:: /url/to/image.png

                ...and some text below.
            """,
            html_string="""
                <p>A inline <img alt="substitution text" src="/url/to/image.png" /> image.</p>
                <p>...and some text below.</p>
            """
        )

    def test_table(self):
        self.cross_compare(
            rest_string="""
                before table.

                +------------+
                | table item |
                +------------+

                After table.
            """,
            html_string="""
                <p>before table.</p>
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td>table item</td>
                </tr>
                </table>
                <p>After table.</p>
            """
        )

    def test_link_in_table1(self):
        self.cross_compare(
            rest_string="""
                +---------------+
                | `table item`_ |
                +---------------+

                .. _table item: foo/bar
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td><a href="foo/bar">table item</a></td>
                </tr>
                </table>
            """
        )

    def test_link_in_table2(self):
        self.cross_compare(
            rest_string="""
                +-----------------------+
                | foo `table item`_ bar |
                +-----------------------+

                .. _table item: foo/bar
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td>foo <a href="foo/bar">table item</a> bar</td>
                </tr>
                </table>
            """
        )

    def test_link_in_table3(self):
        self.cross_compare(
            rest_string="""
                +-----------------------------+
                | * foo `table item 1`_ bar 1 |
                +-----------------------------+
                | * foo `table item 2`_ bar 2 |
                +-----------------------------+

                .. _table item 1: foo/bar/1/
                .. _table item 2: foo/bar/2/
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td><ul>
                <li>foo <a href="foo/bar/1/">table item 1</a> bar 1</li>
                </ul>
                </td>
                </tr>
                <tr><td><ul>
                <li>foo <a href="foo/bar/2/">table item 2</a> bar 2</li>
                </ul>
                </td>
                </tr>
                </table>
            """
        )

    def test_paragraph_bwlow_table_links(self):
        self.cross_compare(
            rest_string="""
                +-----------------+
                | `table item 1`_ |
                +-----------------+
                | `table item 2`_ |
                +-----------------+

                .. _table item 1: foo/bar/1/
                .. _table item 2: foo/bar/2/

                Text after table.
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td><a href="foo/bar/1/">table item 1</a></td>
                </tr>
                <tr><td><a href="foo/bar/2/">table item 2</a></td>
                </tr>
                </table>
                <p>Text after table.</p>
            """,
            # debug=True
        )

    def test_reuse_link_substitution1(self):
        self.cross_compare(
            rest_string="""
                +--------------------------------+
                | this is `foo bar`_ first time. |
                +--------------------------------+
                | and here `foo bar`_ again.     |
                +--------------------------------+

                .. _foo bar: foo/bar/

                Text after table.
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td>this is <a href="foo/bar/">foo bar</a> first time.</td>
                </tr>
                <tr><td>and here <a href="foo/bar/">foo bar</a> again.</td>
                </tr>
                </table>
                <p>Text after table.</p>
            """,
            # debug=True
        )

    def test_reuse_link_substitution2(self):
        self.cross_compare(
            rest_string="""
                +--------------------------------+
                | this is `foo bar`_ first time. |
                +--------------------------------+

                .. _foo bar: foo/bar/

                and here `foo bar`_ again, after table.
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td>this is <a href="foo/bar/">foo bar</a> first time.</td>
                </tr>
                </table>
                <p>and here <a href="foo/bar/">foo bar</a> again, after table.</p>
            """,
            # debug=True
        )

    def test_reuse_image_substitution(self):
        self.cross_compare(
            rest_string="""
                +----------------------+
                | first |image| here   |
                +----------------------+
                | second |image| there |
                +----------------------+

                .. |image| image:: /picture.png
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td>first <img alt="image" src="/picture.png" /> here</td>
                </tr>
                <tr><td>second <img alt="image" src="/picture.png" /> there</td>
                </tr>
                </table>
            """,
            # debug=True
        )

    def test_duplicate_image_substitution(self):
        self.cross_compare(
            rest_string="""
                +----------------------+
                | a |same| image here  |
                +----------------------+
                | a `same`_ link there |
                +----------------------+

                .. |same| image:: /image.png
                .. _same: /url/foo/

                again: the |same| image and `same`_ link!
            """,
            html_string="""
                <table>
                <colgroup>
                <col width="100%" />
                </colgroup>
                <tr><td>a <img alt="same" src="/image.png" /> image here</td>
                </tr>
                <tr><td>a <a href="/url/foo/">same</a> link there</td>
                </tr>
                </table>
                <p>again: the <img alt="same" src="/image.png" /> image and <a href="/url/foo/">same</a> link!</p>
            """,
            # debug=True
        )


#    def test_inline_literal(self):
#        """ TODO
#        http://docutils.sourceforge.net/docs/user/rst/quickref.html#inline-markup
#        """
#        self.cross_compare_rest(
#            rest_string="""
#                ``inline literal``
#            """,
#            html_string="""
#                <p><code>inline&nbsp;literal</code></p>
#            """
#        )

#    def test_escape_in_pre(self):
#        self.cross_compare_rest(
#            textile_string="""
#                <pre>
#                <html escaped>
#                </pre>
#            """,
#            html_string="""
#                <pre>
#                &#60;html escaped&#62;
#                </pre>
#            """)
if __name__ == '__main__':
    unittest.main(
        #        defaultTest="CrossCompareReStTests.test_paragraph_bwlow_table_links",
    )



"""
    :copyleft: 2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from pathlib import Path

from poetry_publish.tests.test_project_setup import test_poetry_check as assert_poetry_check
from poetry_publish.tests.test_project_setup import test_version as assert_version

from creole import __version__
from creole.setup_utils import update_rst_readme
from creole.tests.constants import CREOLE_PACKAGE_ROOT


def test_version():
    """
    Check if current version exists in README
    Check if current version is in pyproject.toml
    """
    assert_version(package_root=CREOLE_PACKAGE_ROOT, version=__version__)


def test_update_rst_readme(capsys):
    rest_readme_path = update_rst_readme(
        package_root=CREOLE_PACKAGE_ROOT, filename='README.creole'
    )
    captured = capsys.readouterr()
    assert captured.out == 'Generate README.rst from README.creole...nothing changed, ok.\n'
    assert captured.err == ''
    assert isinstance(rest_readme_path, Path)
    assert str(rest_readme_path).endswith('/README.rst')


def test_poetry_check():
    """
    Test 'poetry check' output.
    """
    assert_poetry_check(package_root=CREOLE_PACKAGE_ROOT)



"""
    A clean reStructuredText html writer
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    It will produce a minimal set of html output.
    (No extry divs, classes oder ids.)

    Some code stolen from:
    http://www.arnebrodowski.de/blog/write-your-own-restructuredtext-writer.html
    https://github.com/alex-morega/docutils-plainhtml/blob/master/plain_html_writer.py

    :copyleft: 2011-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import sys

from creole.exceptions import DocutilsImportError


try:
    import docutils
    from docutils.core import publish_parts
    from docutils.writers import html4css1
except ImportError:
    etype, evalue, etb = sys.exc_info()
    msg = (
        "%s - You can't use rest2html!"
        " Please install: http://pypi.python.org/pypi/docutils"
    ) % evalue
    evalue = etype(msg)

    # Doesn't work with Python 3:
    # http://www.python-forum.de/viewtopic.php?f=1&t=27507
    # raise DocutilsImportError, evalue, etb

    raise DocutilsImportError(msg)


DEBUG = False
#DEBUG = True

IGNORE_ATTR = (
    "start", "class", "frame", "rules",
)
IGNORE_TAGS = (
    "div",
)


class CleanHTMLWriter(html4css1.Writer):
    """
    This docutils writer will use the CleanHTMLTranslator class below.
    """

    def __init__(self):
        html4css1.Writer.__init__(self)
        self.translator_class = CleanHTMLTranslator


class CleanHTMLTranslator(html4css1.HTMLTranslator):
    """
    Clean html translator for docutils system.
    """

    def _do_nothing(self, node, *args, **kwargs):
        pass

    def starttag(self, node, tagname, suffix='\n', empty=0, **attributes):
        """
        create start tag with the filter IGNORE_TAGS and IGNORE_ATTR.
        """
#        return super(CleanHTMLTranslator, self).starttag(node, tagname, suffix, empty, **attributes)
#        return "XXX%r" % tagname

        if tagname in IGNORE_TAGS:
            if DEBUG:
                print(f"ignore tag {tagname!r}")
            return ""

        parts = [tagname]
        for name, value in sorted(attributes.items()):
            # value=None was used for boolean attributes without
            # value, but this isn't supported by XHTML.
            assert value is not None

            name = name.lower()

            if name in IGNORE_ATTR:
                continue

            if isinstance(value, list):
                value = ' '.join([str(x) for x in value])

            part = f'{name.lower()}="{self.attval(str(value))}"'
            parts.append(part)

        if DEBUG:
            print(
                f'Tag {tagname!r}'
                f' - ids: {getattr(node, "ids", "-")!r}'
                f' - attributes: {attributes!r}'
                f' - parts: {parts!r}'
            )

        if empty:
            infix = ' /'
        else:
            infix = ''
        html = f"<{' '.join(parts)}{infix}>{suffix}"
        if DEBUG:
            print(f"startag html: {html!r}")
        return html

    def visit_section(self, node):
        self.section_level += 1

    def depart_section(self, node):
        self.section_level -= 1

    set_class_on_child = _do_nothing
    set_first_last = _do_nothing

    # remove <blockquote> (e.g. in nested lists)
    visit_block_quote = _do_nothing
    depart_block_quote = _do_nothing

    # set only html_body, we used in rest2html() and don't surround it with <div>
    def depart_document(self, node):
        self.html_body.extend(
            self.body_prefix[1:] + self.body_pre_docinfo + self.docinfo + self.body + self.body_suffix[:-1]
        )
        assert not self.context, f'len(context) = {len(self.context)}'

    # __________________________________________________________________________
    # Clean table:

    visit_thead = _do_nothing
    depart_thead = _do_nothing
    visit_tbody = _do_nothing
    depart_tbody = _do_nothing

    def visit_table(self, node):
        if docutils.__version__ > "0.10":
            self.context.append(self.compact_p)
            self.compact_p = True
        self.body.append(self.starttag(node, 'table'))

    def visit_tgroup(self, node):
        node.stubs = []

    def visit_field_list(self, node):
        super().visit_field_list(node)
        if "<col" in self.body[-1]:
            del(self.body[-1])

    def depart_field_list(self, node):
        self.body.append('</table>\n')
        self.compact_field_list, self.compact_p = self.context.pop()

    def visit_docinfo(self, node):
        self.body.append(self.starttag(node, 'table'))

    def depart_docinfo(self, node):
        self.body.append('</table>\n')

    # __________________________________________________________________________
    # Clean image:

    depart_figure = _do_nothing

    def visit_image(self, node):
        super().visit_image(node)
        if self.body[-1].startswith('<img'):
            align = None

            if 'align' in node:
                # image with alignment
                align = node['align']

            elif node.parent.tagname == 'figure' and 'align' in node.parent:
                # figure with alignment
                align = node.parent['align']

            if align:
                self.body[-1] = self.body[-1].replace(' />', f' align="{align}" />')


def rest2html(content, enable_exit_status=None, **kwargs):
    """
    Convert reStructuredText markup to clean html code: No extra div, class or ids.

    >>> rest2html("- bullet list")
    '<ul>\\n<li>bullet list</li>\\n</ul>\\n'

    >>> rest2html("A ReSt link to `PyLucid CMS <http://www.pylucid.org>`_ :)")
    '<p>A ReSt link to <a href="http://www.pylucid.org">PyLucid CMS</a> :)</p>\\n'

    >>> rest2html("========", enable_exit_status=1, traceback=False, exit_status_level=2)
    Traceback (most recent call last):
    ...
    SystemExit: 13
    """
    assert isinstance(content, str), f"rest2html content must be {str}, but it's {type(content)}"

    settings_overrides = {
        "input_encoding": "unicode",
        "doctitle_xform": False,
        "file_insertion_enabled": False,
        "raw_enabled": False,
    }
    settings_overrides.update(kwargs)

    parts = publish_parts(
        source=content,
        writer=CleanHTMLWriter(),
        settings_overrides=settings_overrides,
        enable_exit_status=enable_exit_status,
    )
#    import pprint
#    pprint.pprint(parts)
    return parts["html_body"]  # Don't detache the first heading


if __name__ == '__main__':
    import doctest
    print(doctest.testmod())

#    print(rest2html(""")
# +------------+------------+
# | Headline 1 | Headline 2 |
# +============+============+
# | cell one   | cell two   |
# +------------+------------+
#    """)

#    print(rest2html(""")
#:homepage:
#  http://code.google.com/p/python-creole/
#
#:sourcecode:
#  http://github.com/jedie/python-creole
#    """)

    print(rest2html("""
===============
Section Title 1
===============

---------------
Section Title 2
---------------

Section Title 3
===============

Section Title 4
---------------

Section Title 5
```````````````

Section Title 6
'''''''''''''''
    """))



"""
    WikiCreole to HTML converter

    :copyleft: 2008-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import json
import sys
import traceback
from xml.sax.saxutils import escape

from creole.parser.creol2html_parser import CreoleParser
from creole.shared.utils import string2dict


class TableOfContent:
    def __init__(self):
        self.max_depth = None
        self.headlines = []
        self._created = False
        self._current_level = 0

    def __call__(self, depth=None, **kwargs):
        """Called when if the macro <<toc>> is defined when it is emitted."""
        if self._created:
            return "&lt;&lt;toc&gt;&gt;"

        self._created = True
        if depth is not None:
            self.max_depth = depth

        return '<<toc>>'

    def add_headline(self, level, content):
        """Add the current header to the toc."""
        if self.max_depth is None or level <= self.max_depth:
            self.headlines.append(
                (level, content)
            )

    def flat_list2nest_list(self, flat_list):
        # this func code based on borrowed code from EyDu, Thanks!
        # http://www.python-forum.de/viewtopic.php?p=258121#p258121
        tree = []
        stack = [tree]

        for index, element in flat_list:
            stack_length = len(stack)

            if index > stack_length:
                for _ in range(stack_length, index):
                    nest_list = []
                    stack[-1].append(nest_list)
                    stack.append(nest_list)
            elif index < stack_length:
                stack = stack[:index]

            stack[-1].append(element)

        return tree

    def nested_headlines2html(self, nested_headlines, level=0):
        """Convert a python nested list like the one representing the toc to an html equivalent."""
        indent = "\t" * level
        if isinstance(nested_headlines, str):
            return f'{indent}<li><a href="#{nested_headlines}">{nested_headlines}</a></li>\n'
        elif isinstance(nested_headlines, list):
            html = f'{indent}<ul>\n'
            for elt in nested_headlines:
                html += self.nested_headlines2html(elt, level + 1)
            html += f'{indent}</ul>'
            if level > 0:
                html += "\n"
            return html

    def emit(self, document):
        """Emit the toc where the <<toc>> macro was."""
        nested_headlines = self.flat_list2nest_list(self.headlines)
        html = self.nested_headlines2html(nested_headlines)

        # FIXME: We should not use <p> here, because it doesn't match
        #        if no newline was made before <<toc>>
        if "<p><<toc>></p>" in document:
            document = document.replace("<p><<toc>></p>", html, 1)
        else:
            document = document.replace("<<toc>>", html, 1)

        return document


class HtmlEmitter:
    """
    Generate HTML output for the document
    tree consisting of DocNodes.
    """

    def __init__(self, root, macros=None, verbose=None, stderr=None, strict=False):

        self.root = root

        if callable(macros):
            # was a DeprecationWarning in the past
            raise TypeError("Callable macros are not supported anymore!")

        if macros is None:
            self.macros = {}
        else:
            self.macros = macros

        if "toc" not in root.used_macros:
            # The document has no <<toc>>
            self.toc = None
        else:
            if isinstance(self.macros, dict):
                if "toc" in self.macros:
                    self.toc = self.macros["toc"]
                else:
                    self.toc = TableOfContent()
                    self.macros["toc"] = self.toc
            else:
                try:
                    self.toc = getattr(self.macros, "toc")
                except AttributeError:
                    self.toc = TableOfContent()
                    self.macros.toc = self.toc

        if verbose is None:
            self.verbose = 1
        else:
            self.verbose = verbose

        if stderr is None:
            self.stderr = sys.stderr
        else:
            self.stderr = stderr

        self.strict = strict

    def get_text(self, node):
        """Try to emit whatever text is in the node."""
        try:
            return node.children[0].content or ''
        except BaseException:
            return node.content or ''

    def html_escape(self, text):
        return escape(text)

    def attr_escape(self, text):
        return self.html_escape(text).replace('"', '&quot')

    # *_emit methods for emitting nodes of the document:

    def document_emit(self, node):
        return self.emit_children(node)

    def text_emit(self, node):
        return self.html_escape(node.content)

    def separator_emit(self, node):
        return '<hr />\n'

    def paragraph_emit(self, node):
        return f'<p>{self.emit_children(node)}</p>\n'

    def _list_emit(self, node, list_type):
        if node.parent.kind in ("document",):
            # The first list item
            formatter = ''
        else:
            formatter = '\n'

        if list_type == "li":
            formatter += (
                '%(i)s<%(t)s>%(c)s</%(t)s>'
            )
        else:
            formatter += (
                '%(i)s<%(t)s>%(c)s\n'
                '%(i)s</%(t)s>'
            )
        return formatter % {
            "i": "\t" * node.level,
            "c": self.emit_children(node),
            "t": list_type,
        }

    def bullet_list_emit(self, node):
        return self._list_emit(node, list_type="ul")

    def number_list_emit(self, node):
        return self._list_emit(node, list_type="ol")

    def list_item_emit(self, node):
        return self._list_emit(node, list_type="li")

    def table_emit(self, node):
        return f'<table>\n{self.emit_children(node)}</table>\n'

    def table_row_emit(self, node):
        return f'<tr>\n{self.emit_children(node)}</tr>\n'

    def table_cell_emit(self, node):
        return f'\t<td>{self.emit_children(node)}</td>\n'

    def table_head_emit(self, node):
        return f'\t<th>{self.emit_children(node)}</th>\n'

    # --------------------------------------------------------------------------

    def _typeface(self, node, tag):
        return '<{tag}>{data}</{tag}>'.format(
            tag=tag,
            data=self.emit_children(node),
        )

    # TODO: How can we generalize that:
    def emphasis_emit(self, node):
        return self._typeface(node, tag="i")

    def strong_emit(self, node):
        return self._typeface(node, tag="strong")

    def monospace_emit(self, node):
        return self._typeface(node, tag="tt")

    def superscript_emit(self, node):
        return self._typeface(node, tag="sup")

    def subscript_emit(self, node):
        return self._typeface(node, tag="sub")

    def underline_emit(self, node):
        return self._typeface(node, tag="u")

    def small_emit(self, node):
        return self._typeface(node, tag="small")

    def delete_emit(self, node):
        return self._typeface(node, tag="del")

    # --------------------------------------------------------------------------

    def header_emit(self, node):
        header = f'<h{node.level:d}>{self.html_escape(node.content)}</h{node.level:d}>'
        if self.toc is not None:
            self.toc.add_headline(node.level, node.content)
            # add link attribute for toc navigation
            header = f'<a name="{self.html_escape(node.content)}">{header}</a>'

        header += "\n"
        return header

    def preformatted_emit(self, node):
        return f'<pre>{self.html_escape(node.content)}</pre>'

    def link_emit(self, node):
        target = node.content
        if node.children:
            inside = self.emit_children(node)
        else:
            inside = self.html_escape(target)

        return f'<a href="{self.attr_escape(target)}">{inside}</a>'

    def image_emit(self, node):
        target = node.content
        text = self.attr_escape(self.get_text(node))
        if not self.strict:
            try:
                if "|" in text:
                    title, size_str = text.split("|", 1)
                    if not title:
                        title = target
                    w_str, h_str = size_str.split("x", 1)
                    width = int(w_str.strip())
                    height = int(h_str.strip())
                    return f'<img src="{self.attr_escape(target)}" title="{title}" alt="{title}" width="{width}" height="{height}" />'
            except BaseException:
                pass
        return f'<img src="{self.attr_escape(target)}" title="{text}" alt="{text}" />'

    def macro_emit(self, node):
        # print(node.debug())
        macro_name = node.macro_name
        text = node.content
        macro = None

        args = node.macro_args
        try:
            macro_kwargs = string2dict(args)
        except ValueError:
            exc_info = sys.exc_info()
            return self.error(
                f"Wrong macro arguments: {json.dumps(args)} for macro '{macro_name}' (maybe wrong macro tag syntax?)",
                exc_info
            )

        macro_kwargs["text"] = text

        exc_info = None
        if isinstance(self.macros, dict):
            try:
                macro = self.macros[macro_name]
            except KeyError:
                exc_info = sys.exc_info()
        else:
            try:
                macro = getattr(self.macros, macro_name)
            except AttributeError:
                exc_info = sys.exc_info()

        if macro is None:
            return self.error(
                f"Macro '{macro_name}' doesn't exist",
                exc_info
            )

        try:
            result = macro(**macro_kwargs)
        except TypeError as err:
            msg = f"Macro '{macro_name}' error: {err}"
            exc_info = sys.exc_info()
            if self.verbose > 1:
                if self.verbose > 2:
                    raise

                # Inject more information about the macro in traceback
                etype, evalue, etb = exc_info
                import inspect
                try:
                    filename = inspect.getfile(macro)
                except TypeError:
                    pass
                else:
                    try:
                        sourceline = inspect.getsourcelines(macro)[0][0].strip()
                    except OSError as err:
                        evalue = etype(f"{evalue} (error getting sourceline: {err} from {filename})")
                    else:
                        evalue = etype(f"{evalue} (sourceline: {sourceline!r} from {filename})")
                    exc_info = etype, evalue, etb

            return self.error(msg, exc_info)
        except Exception as err:
            return self.error(
                f"Macro '{macro_name}' error: {err}",
                exc_info=sys.exc_info()
            )

        if not isinstance(result, str):
            msg = f"Macro '{macro_name}' doesn't return a unicode string!"
            if self.verbose > 1:
                msg += f" - returns: {result!r}, type {type(result)!r}"
            return self.error(msg)

        if node.kind == "macro_block":
            result += "\n"

        return result
    macro_inline_emit = macro_emit
    macro_block_emit = macro_emit

    def break_emit(self, node):
        if node.parent.kind == "list_item":
            return "<br />\n" + "\t" * node.parent.level
        elif node.parent.kind in ("table_head", "table_cell"):
            return "<br />\n\t\t"
        else:
            return "<br />\n"

    def line_emit(self, node):
        return "\n"

    def pre_block_emit(self, node):
        """ pre block, with newline at the end """
        return f"<pre>{self.html_escape(node.content)}</pre>\n"

    def pre_inline_emit(self, node):
        """ pre without newline at the end """
        return f"<tt>{self.html_escape(node.content)}</tt>"

    def default_emit(self, node):
        """Fallback function for emitting unknown nodes."""
        raise NotImplementedError(f"Node '{node.kind}' unknown")

    def emit_children(self, node):
        """Emit all the children of a node."""
        return ''.join([self.emit_node(child) for child in node.children])

    def emit_node(self, node):
        """Emit a single node."""
        #print("%s_emit: %r" % (node.kind, node.content))
        emit = getattr(self, f'{node.kind}_emit', self.default_emit)
        return emit(node)

    def emit(self):
        """Emit the document represented by self.root DOM tree."""
        document = self.emit_node(self.root).strip()
        if self.toc is not None:
            return self.toc.emit(document)
        else:
            return document

    def error(self, text, exc_info=None):
        """
        Error Handling.
        """
        if self.verbose > 1 and exc_info:
            exc_type, exc_value, exc_traceback = exc_info
            exception = "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            self.stderr.write(exception)

        if self.verbose > 0:
            return f"[Error: {text}]\n"
        else:
            # No error output
            return ""


if __name__ == "__main__":
    txt = """Local test
<<toc>>
= headline 1 level 1
== headline 2 level 2
== headline 3 level 2
==== headline 4 level 4
= headline 5 level 1
=== headline 6 level 3
"""

    print("-" * 80)
#    from creole_alt.creole import CreoleParser
    p = CreoleParser(txt)
    document = p.parse()
    p.debug()

    html = HtmlEmitter(document, verbose=999).emit()
    print(html)
    print("-" * 79)
    print(html.replace(" ", ".").replace("\n", "\\n\n"))



"""
    html2rest unittest
    ~~~~~~~~~~~~~~~~~~~~~

    Unittests for special cases which only works in the html2rest way.

    Note: This only works fine if there is no problematic whitespace handling.

    :copyleft: 2011-2012 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import unittest

from creole.emitter.html2rest_emitter import Html2restException
from creole.tests.utils.base_unittest import BaseCreoleTest


class ReStTests(BaseCreoleTest):
    def test_line_breaks(self):
        """
        Line breaks in HTML are lost.
        """
        self.assert_html2rest(
            rest_string="""
                first block, line 1 and line 2

                second block, line 1 and line 2
            """,
            html_string="""
                <p>first block, line 1
                and line 2</p>
                <p>second block, line 1
                and line 2</p>
            """,
            # debug=True
        )

    def test_substitution_image_without_alt_or_title(self):
        self.assert_html2rest(
            rest_string="""
                A inline |image.png| image.

                .. |image.png| image:: /url/to/image.png

                ...and some text below.
            """,
            html_string="""
                <p>A inline <img src="/url/to/image.png" /> image.</p>
                <p>...and some text below.</p>
            """
        )

    def test_substitution_image_with_title(self):
        self.assert_html2rest(
            rest_string="""
                A inline |foo bar| image.

                .. |foo bar| image:: /url/to/image.png

                ...and some text below.
            """,
            html_string="""
                <p>A inline <img title="foo bar" src="/url/to/image.png" /> image.</p>
                <p>...and some text below.</p>
            """
        )

    def test_substitution_image_without_p(self):
        self.assert_html2rest(
            rest_string="""
                |image.png|

                .. |image.png| image:: /url/to/image.png
            """,
            html_string="""
                <img src="/url/to/image.png" />
            """
        )

    def test_pre_code1(self):
        self.assert_html2rest(
            rest_string="""
                Text line

                ::

                    >>> from creole import creole2html
                    >>> creole2html("This is **creole //markup//**")
                    '<p>This is <strong>creole <i>markup</i></strong></p>'
            """,
            html_string="""
                <p>Text line</p>
                <pre>
                &gt;&gt;&gt; from creole import creole2html
                &gt;&gt;&gt; creole2html(&quot;This is **creole //markup//**&quot;)
                '&lt;p&gt;This is &lt;strong&gt;creole &lt;i&gt;markup&lt;/i&gt;&lt;/strong&gt;&lt;/p&gt;'
                </pre>
            """
        )

    def test_escape(self):
        self.assert_html2rest(
            rest_string="""
                * Use <tt> when {{{ ... }}} is inline and not <pre>, or not?
            """,
            html_string="""
                <ul>
                <li>Use &lt;tt&gt; when {{{ ... }}} is inline and not &lt;pre&gt;, or not?</li>
                </ul>
            """
        )

    def test_inline_literals(self):
        self.assert_html2rest(
            rest_string="""
                This text is an example of ``inline literals``.
            """,
            html_string="""
                <ul>
                <p>This text is an example of <tt>inline literals</tt>.</p>
                </ul>
            """
        )

    def test_list_without_p(self):
        self.assert_html2rest(
            rest_string="""
                A nested bullet lists:

                * item 1 without p-tag

                    * A **`subitem 1.1 </1.1/url/>`_ link** here.

                        * subsubitem 1.1.1

                        * subsubitem 1.1.2

                    * subitem 1.2

                * item 2 without p-tag

                    * subitem 2.1

                Text under list.
            """,
            html_string="""
                <p>A nested bullet lists:</p>
                <ul>
                    <li>item 1 without p-tag
                        <ul>
                            <li>A <strong><a href="/1.1/url/">subitem 1.1</a> link</strong> here.
                                <ul>
                                    <li>subsubitem 1.1.1</li>
                                    <li>subsubitem 1.1.2</li>
                                </ul>
                            </li>
                            <li>subitem 1.2</li>
                        </ul>
                    </li>
                    <li>item 2 without p-tag
                        <ul>
                            <li>subitem 2.1</li>
                        </ul>
                    </li>
                </ul>
                <p>Text under list.</p>
            """
        )

    def test_table_with_headings(self):
        self.assert_html2rest(
            rest_string="""
                +--------+--------+
                | head 1 | head 2 |
                +========+========+
                | item 1 | item 2 |
                +--------+--------+
            """,
            html_string="""
                <table>
                <tr><th>head 1</th><th>head 2</th>
                </tr>
                <tr><td>item 1</td><td>item 2</td>
                </tr>
                </table>
            """
        )

    def test_table_without_headings(self):
        self.assert_html2rest(
            rest_string="""
                +--------+--------+
                | item 1 | item 2 |
                +--------+--------+
                | item 3 | item 4 |
                +--------+--------+
            """,
            html_string="""
                <table>
                <tr><td>item 1</td><td>item 2</td>
                </tr>
                <tr><td>item 3</td><td>item 4</td>
                </tr>
                </table>
            """
        )

    def test_duplicate_substitution1(self):
        self.assertRaises(Html2restException, self.assert_html2rest,
                          rest_string="""
                +-----------------------------+
                | this is `same`_ first time. |
                +-----------------------------+

                .. _same: /first/

                the `same </other/>`_ link?
            """,
                          html_string="""
                <table>
                <tr><td>the <a href="/first/">same</a> first time.</td>
                </tr>
                </table>
                <p>the <a href="/other/">same</a> link?</p>
            """,
                          # debug=True
                          )

    def test_duplicate_link_substitution(self):
        self.assertRaises(Html2restException, self.assert_html2rest,
                          #        self.cross_compare(
                          rest_string="""
                +-----------------------------+
                | this is `same`_ first time. |
                +-----------------------------+

                .. _same: /first/

                the `same </other/>`_ link?
            """,
                          html_string="""
                <table>
                <tr><td>the <a href="/first/">same</a> first time.</td>
                </tr>
                </table>
                <p>the <a href="/other/">same</a> link?</p>
            """,
                          # debug=True
                          )

    def test_duplicate_image_substitution(self):
        self.assertRaises(Html2restException, self.assert_html2rest,
                          #        self.cross_compare(
                          rest_string="""
                a |image|...
                and a other |image|!

                .. |image| image:: /image.png
                .. |image| image:: /other.png
            """,
                          html_string="""
                <p>a <img src="/image.png" title="image" alt="image" />...<br />
                and a other <img src="/other.png" title="image" alt="image" />!</p>
            """,
                          # debug=True
                          )


#    def test_preformat_unknown_nodes(self):
#        """
#        Put unknown tags in a <pre> area.
#        """
#        self.assert_html2rest(
#            rest_string="""
#                111 <<pre>><x><</pre>>foo<<pre>></x><</pre>> 222
#                333<<pre>><x foo1="bar1"><</pre>>foobar<<pre>></x><</pre>>444
#
#                555<<pre>><x /><</pre>>666
#            """,
#            html_string="""
#                <p>111 <x>foo</x> 222<br />
#                333<x foo1="bar1">foobar</x>444</p>
#
#                <p>555<x />666</p>
#            """,
#            emitter_kwargs={"unknown_emit":preformat_unknown_nodes}
#        )
#
#    def test_transparent_unknown_nodes(self):
#        """
#        transparent_unknown_nodes is the default unknown_emit:
#
#        Remove all unknown html tags and show only
#        their child nodes' content.
#        """
#        self.assert_html2rest(
#            rest_string="""
#                111 <<pre>><x><</pre>>foo<<pre>></x><</pre>> 222
#                333<<pre>><x foo1="bar1"><</pre>>foobar<<pre>></x><</pre>>444
#
#                555<<pre>><x /><</pre>>666
#            """,
#            html_string="""
#                <p>111 <x>foo</x> 222<br />
#                333<x foo1="bar1">foobar</x>444</p>
#
#                <p>555<x />666</p>
#            """,
#        )

if __name__ == '__main__':
    unittest.main()



"""
    :copyleft: 2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import difflib


def unified_diff(old, new, filename=None):
    """
    Return text of unified diff between old and new.
    """
    if filename is None:
        fromfile = 'old'
        tofile = 'new'
    else:
        fromfile = f'old / {filename}'
        tofile = f'new / {filename}'

    if isinstance(old, str) and isinstance(new, str):
        old = old.splitlines(keepends=True)
        new = new.splitlines(keepends=True)

    diff = difflib.unified_diff(old, new, fromfile=fromfile, tofile=tofile)

    text = ''
    for line in diff:
        text += line

        # Work around missing newline (http://bugs.python.org/issue2142).
        if text and not line.endswith('\n'):
            text += 'n\\ No newline at end of file\n'

    return text



import tempfile
from pathlib import Path

import pytest

from creole.setup_utils import _generate_rst_readme, update_rst_readme
from creole.tests.constants import CREOLE_PACKAGE_ROOT


def test_update_rst_readme(capsys):
    rest_readme_path = update_rst_readme(
        package_root=CREOLE_PACKAGE_ROOT, filename='README.creole'
    )
    captured = capsys.readouterr()
    assert captured.out == 'Generate README.rst from README.creole...nothing changed, ok.\n'
    assert captured.err == ''
    assert isinstance(rest_readme_path, Path)
    assert str(rest_readme_path).endswith('/README.rst')


def test_non_valid_readme(capsys):
    with tempfile.NamedTemporaryFile() as fp:
        path = Path(fp.name)
        with path.open('w') as f:
            f.write('= headline\n')
            f.write('\n')
            f.write('----\n')  # << error ;)

        with pytest.raises(SystemExit):
            _generate_rst_readme(creole_readme_path=path)

        captured = capsys.readouterr()
        assert captured.out == ''
        assert captured.err == (
            '<string>:5: (ERROR/3) Document or section may not begin with a transition.\n'
        )



"""
    html -> reStructuredText Emitter
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Links about reStructuredText:

    http://openalea.gforge.inria.fr/doc/openalea/doc/_build/html/source/sphinx/rest_syntax.html

    :copyleft: 2011-2012 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import posixpath

from creole.shared.base_emitter import BaseEmitter
from creole.shared.markup_table import MarkupTable


# Kink of nodes in which hyperlinks are stored in references intead of embedded urls.
DO_SUBSTITUTION = ("th", "td",)  # TODO: In witch kind of node must we also substitude links?


class Html2restException(Exception):
    pass


class ReStructuredTextEmitter(BaseEmitter):
    """
    Build from a document_tree (html2creole.parser.HtmlParser instance) a
    creole markup text.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.table_head_prefix = "_. "
        self.table_auto_width = False

        self._substitution_data = []
        self._used_substitution_links = {}
        self._used_substitution_images = {}
        self._list_markup = ""

    def _get_block_data(self):
        """
        return substitution bock data
        e.g.:
        .. _link text: /link/url/
        .. |substitution| image:: /image.png
        """
        content = "\n".join(self._substitution_data)
        self._substitution_data = []
        return content

    # --------------------------------------------------------------------------

    def blockdata_pre_emit(self, node):
        """ pre block -> with newline at the end """
        pre_block = self.deentity.replace_all(node.content).strip()
        pre_block = "\n".join(["    %s" % line for line in pre_block.splitlines()])
        return f"::\n\n{pre_block}\n\n"

    def inlinedata_pre_emit(self, node):
        """ a pre inline block -> no newline at the end """
        return f"<pre>{self.deentity.replace_all(node.content)}</pre>"

    def blockdata_pass_emit(self, node):
        return f"{node.content}\n\n"

    # --------------------------------------------------------------------------

    def emit_children(self, node):
        """Emit all the children of a node."""
        return "".join(self.emit_children_list(node))

    def emit(self):
        """Emit the document represented by self.root DOM tree."""
        return self.emit_node(self.root).rstrip()

    def document_emit(self, node):
        self.last = node
        result = self.emit_children(node)
        if self._substitution_data:
            # add rest at the end
            if not result.endswith("\n\n"):
                result += "\n\n"
            result += f"{self._get_block_data()}\n\n"
        return result

    def emit_node(self, node):
        result = ""
        if self._substitution_data and node.parent == self.root:
            result += f"{self._get_block_data()}\n\n"

        result += super().emit_node(node)
        return result

    def p_emit(self, node):
        return f"{self.emit_children(node)}\n\n"

    HEADLINE_DATA = {
        1: ("=", True),
        2: ("-", True),
        3: ("=", False),
        4: ("-", False),
        5: ('`', False),
        6: ("'", False),
    }

    def headline_emit(self, node):
        text = self.emit_children(node)

        level = node.level
        if level > 6:
            level = 6

        char, both = self.HEADLINE_DATA[level]
        markup = char * len(text)

        if both:
            format = "%(m)s\n%(t)s\n%(m)s\n\n"
        else:
            format = "%(t)s\n%(m)s\n\n"

        return format % {"m": markup, "t": text}

    # --------------------------------------------------------------------------

    def _typeface(self, node, key):
        return key + self.emit_children(node) + key

    def strong_emit(self, node):
        return self._typeface(node, key="**")

    def b_emit(self, node):
        return self._typeface(node, key="**")
    big_emit = strong_emit

    def i_emit(self, node):
        return self._typeface(node, key="*")

    def em_emit(self, node):
        return self._typeface(node, key="*")

    def tt_emit(self, node):
        return self._typeface(node, key="``")

    def small_emit(self, node):
        # FIXME: Is there no small in ReSt???
        return self.emit_children(node)

#    def sup_emit(self, node):
#        return self._typeface(node, key="^")
#    def sub_emit(self, node):
#        return self._typeface(node, key="~")
#    def del_emit(self, node):
#        return self._typeface(node, key="-")
#
#    def cite_emit(self, node):
#        return self._typeface(node, key="??")
#    def ins_emit(self, node):
#        return self._typeface(node, key="+")
#
#    def span_emit(self, node):
#        return self._typeface(node, key="%")
#    def code_emit(self, node):
#        return self._typeface(node, key="@")

    # --------------------------------------------------------------------------

    def hr_emit(self, node):
        return "----\n\n"

    def _should_do_substitution(self, node):
        node = node.parent

        if node.kind in DO_SUBSTITUTION:
            return True

        if node is not self.root:
            return self._should_do_substitution(node)
        else:
            return False

    def _get_old_substitution(self, substitution_dict, text, url):
        if text not in substitution_dict:
            # save for the next time
            substitution_dict[text] = url
        else:
            # text has links with the same link text
            old_url = substitution_dict[text]
            if old_url == url:
                # same url -> substitution can be reused
                return old_url
            else:
                msg = (
                    "Duplicate explicit target name:"
                    " substitution was used more than one time, but with different URL."
                    " - link text: %r url1: %r url2: %r"
                ) % (text, old_url, url)
                raise Html2restException(msg)

    def a_emit(self, node):
        link_text = self.emit_children(node)
        url = node.attrs["href"]

        old_url = self._get_old_substitution(self._used_substitution_links, link_text, url)

        if self._should_do_substitution(node):
            # make a hyperlink reference
            if not old_url:
                # new substitution
                self._substitution_data.append(
                    f".. _{link_text}: {url}"
                )
            return f"`{link_text}`_"

        if old_url:
            # reuse a existing substitution
            return f"`{link_text}`_"
        else:
            # create a inline hyperlink
            return f"`{link_text} <{url}>`_"

    def img_emit(self, node):
        src = node.attrs["src"]

        if src.split(':')[0] == 'data':
            return ""

        title = node.attrs.get("title", "")
        alt = node.attrs.get("alt", "")
        if len(alt) > len(title):  # Use the longest one
            substitution_text = alt
        else:
            substitution_text = title

        if substitution_text == "":  # Use filename as picture text
            substitution_text = posixpath.basename(src)

        old_src = self._get_old_substitution(
            self._used_substitution_images, substitution_text, src
        )
        if not old_src:
            self._substitution_data.append(
                f".. |{substitution_text}| image:: {src}"
            )

        return f"|{substitution_text}|"

    # --------------------------------------------------------------------------

    def code_emit(self, node):
        return f"``{self._emit_content(node)}``"

    # --------------------------------------------------------------------------

    def li_emit(self, node):
        content = self.emit_children(node).strip("\n")
        result = f"\n{'    ' * (node.level - 1)}{self._list_markup} {content}\n"
        return result

    def _list_emit(self, node, list_type):
        self._list_markup = list_type
        content = self.emit_children(node)

        if node.level == 1:
            # FIXME: This should be made ​​easier and better
            complete_list = "\n\n".join([i.strip("\n") for i in content.split("\n") if i])
            content = f"{complete_list}\n\n"

        return content

    def ul_emit(self, node):
        return self._list_emit(node, "*")

    def ol_emit(self, node):
        return self._list_emit(node, "#.")

    def table_emit(self, node):
        """
        http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#tables
        """
        self._table = MarkupTable(
            head_prefix="",
            auto_width=True,
            debug_msg=self.debug_msg
        )
        self.emit_children(node)
        content = self._table.get_rest_table()
        return f"{content}\n\n"



"""
    cross compare textile unittest
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Compare all similarities between:
        * textile2html (used the python textile module)
        * html2textile

    Note: This only works fine if there is no problematic whitespace handling.
        In this case, we must test in test_creole2html.py or test_html2creole.py

    :copyleft: 2008-2014 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import unittest

from creole.tests.utils.base_unittest import BaseCreoleTest


class CrossCompareTextileTests(BaseCreoleTest):
    def test_typeface_basic(self):
        self.cross_compare_textile(
            textile_string="""
                _emphasis_
                *strong*
                __italic__
                **bold**
                ??citation??
                -deleted text-
                +inserted text+
                ^superscript^
                ~subscript~
                %span%
                @code@
            """,
            html_string="""
                <p><em>emphasis</em><br />

                <strong>strong</strong><br />

                <i>italic</i><br />

                <b>bold</b><br />

                <cite>citation</cite><br />

                <del>deleted text</del><br />

                <ins>inserted text</ins><br />

                <sup>superscript</sup><br />

                <sub>subscript</sub><br />

                <span>span</span><br />

                <code>code</code></p>
            """
        )

    def test_escape_in_pre(self):
        self.cross_compare_textile(
            textile_string="""
                <pre>
                <html escaped>
                </pre>
            """,
            html_string="""
                <pre>
                &lt;html escaped&gt;
                </pre>
            """)


if __name__ == '__main__':
    unittest.main()



"""
    html2textile unittest
    ~~~~~~~~~~~~~~~~~~~~~

    Unittests for special cases which only works in the html2textile way.

    Note: This only works fine if there is no problematic whitespace handling.

    :copyleft: 2011 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import unittest

from creole.shared.unknown_tags import preformat_unknown_nodes
from creole.tests.utils.base_unittest import BaseCreoleTest


class TextileTests(BaseCreoleTest):
    def test_entities(self):
        """
        can't be cross tested, because textile would convert < to &#60; and > to &#62;
        """
        self.assert_html2textile(
            textile_string="""
                less-than sign: <
                greater-than sign: >
            """,
            html_string="""
                <p>less-than sign: &lt;<br />
                greater-than sign: &gt;</p>
            """,
            # debug=True
        )

    def test_preformat_unknown_nodes(self):
        """
        Put unknown tags in a <pre> area.
        """
        self.assert_html2textile(
            textile_string="""
                111 <<pre>><x><</pre>>foo<<pre>></x><</pre>> 222
                333<<pre>><x foo1="bar1"><</pre>>foobar<<pre>></x><</pre>>444

                555<<pre>><x /><</pre>>666
            """,
            html_string="""
                <p>111 <x>foo</x> 222<br />
                333<x foo1="bar1">foobar</x>444</p>

                <p>555<x />666</p>
            """,
            emitter_kwargs={"unknown_emit": preformat_unknown_nodes}
        )

    def test_transparent_unknown_nodes(self):
        """
        transparent_unknown_nodes is the default unknown_emit:

        Remove all unknown html tags and show only
        their child nodes' content.
        """
        self.assert_html2textile(
            textile_string="""
                111 foo 222
                333foobar444

                555666
            """,
            html_string="""
                <p>111 <x>foo</x> 222<br />
                333<x foo1="bar1">foobar</x>444</p>

                <p>555<x />666</p>
            """,
        )


if __name__ == '__main__':
    unittest.main()



"""
    :copyleft: 2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import unittest

from creole.tests.utils.unittest_subprocess import SubprocessMixin


class MakefileTestCase(unittest.TestCase, SubprocessMixin):

    def test_help(self):
        popen_args, retcode, stdout = self.subprocess(
            popen_args=["make"],
        )
        assert "List all commands" in stdout
        assert "tox" in stdout
        assert "pytest" in stdout
        assert retcode == 0

    def test_check_poetry(self):
        popen_args, retcode, stdout = self.subprocess(
            popen_args=["make", "check-poetry"],
        )
        assert "Found Poetry version 1." in stdout
        assert "ok" in stdout
        assert retcode == 0



"""
    unitest generic utils
    ~~~~~~~~~~~~~~~~~~~~~

    Generic utils useable for a markup test.

    :copyleft: 2008-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import os
import shutil
import tempfile
import textwrap
import unittest
from pathlib import Path

from creole.shared.diff_utils import unified_diff


class MarkupTest(unittest.TestCase):
    """
    Special error class: Try to display markup errors in a better way.
    """

    # error output format:
    # =1 -> via repr()
    # =2 -> raw
    VERBOSE = 1
    #VERBOSE = 2

    def _format_output(self, txt):
        txt = txt.split("\\n")
        if self.VERBOSE == 1:
            txt = "".join(['%s\\n\n' % i for i in txt])
        elif self.VERBOSE == 2:
            txt = "".join(['%s\n' % i for i in txt])
        return txt

    def assertEqual(self, first, second, msg=""):
        if first == second:
            return

        try:
            diff = unified_diff(first, second)
        except AttributeError:
            raise self.failureException(f"{first!r} is not {second!r}")

        print("*" * 100)
        print("---[Output:]-----------------------------------------------------------------------------------------")
        print(first)
        print("---[not equal to:]-----------------------------------------------------------------------------------")
        print(second)
        print("---[diff:]-------------------------------------------------------------------------------------------")
        print(diff)
        print("*" * 100)

        assert first == second, f"{first!r} is not {second!r}"

    def _prepare_text(self, txt):
        """
        prepare the multiline, indentation text.
        """
        # Remove any common leading whitespace from every line
        txt = textwrap.dedent(txt)

        # Strip spaces and every line end and remove the last line ending:
        txt = "\n".join(line.rstrip(" ") for line in txt.splitlines())

        # strip *one* newline at the beginning...
        if txt.startswith("\n"):
            txt = txt[1:]

        return txt


class IsolatedFilesystem:
    """
    Context manager, e.g.:
    with IsolatedFilesystem(prefix="temp_dir_prefix"):
        print("I'm in the temp path here: %s" % Path().cwd())
    """

    def __init__(self, prefix=None):
        super().__init__()

        self.prefix = prefix

    def __enter__(self):
        print(f"Use prefix: {self.prefix!r}")

        self.cwd = Path().cwd()
        self.temp_path = tempfile.mkdtemp(prefix=self.prefix)
        os.chdir(self.temp_path)

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(str(self.cwd))  # str() needed for older python <=3.5
        try:
            shutil.rmtree(self.temp_path)
        except OSError:
            pass



"""
    cross compare unittest
    ~~~~~~~~~~~~~~~~~~~~~~

    Compare all similarities between:
        * creole2html
        * html2creole
        * textile2html (used the python textile module)
        * html2textile

    Note: This only works fine if there is no problematic whitespace handling.
        In this case, we must test in test_creole2html.py or test_html2creole.py

    :copyleft: 2008-2014 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import unittest

from creole.tests.utils.base_unittest import BaseCreoleTest


class CrossCompareTests(BaseCreoleTest):
    """
    Cross compare tests for creol2html _and_ html2creole with the same test
    strings. Used BaseCreoleTest.assertCreole()
    """

    def test_bold_italics(self):
        self.cross_compare(
            creole_string=r"""
                **bold** //italics//
                //italics and **bold**.//
                **bold and //italics//.**
            """,
            html_string="""
                <p><strong>bold</strong> <i>italics</i><br />
                <i>italics and <strong>bold</strong>.</i><br />
                <strong>bold and <i>italics</i>.</strong></p>
            """,
        )
        self.cross_compare(
            textile_string="""
                *bold* __italics__
                __italics and *bold*.__
                *bold and __italics__.*
            """,
            html_string="""
                <p><strong>bold</strong> <i>italics</i><br />

                <i>italics and <strong>bold</strong>.</i><br />

                <strong>bold and <i>italics</i>.</strong></p>
            """,
        )
        # Note: In ReSt inline markup may not be nested.
        self.cross_compare(
            html_string="""
                <p><strong>bold</strong> <em>italics</em></p>
            """,
            rest_string="""
                **bold** *italics*
            """,
        )

    def test_bold_italics2(self):
        self.cross_compare(
            creole_string=r"""
                **//bold italics//**
                //**bold italics**//
                //This is **also** good.//
            """,
            html_string="""
                <p><strong><i>bold italics</i></strong><br />
                <i><strong>bold italics</strong></i><br />
                <i>This is <strong>also</strong> good.</i></p>
            """,
        )
        self.cross_compare(
            textile_string="""
                *__bold italics__*
                __*bold italics*__
                __This is *also* good.__
            """,
            html_string="""
                <p><strong><i>bold italics</i></strong><br />

                <i><strong>bold italics</strong></i><br />

                <i>This is <strong>also</strong> good.</i></p>
            """,
        )

    def test_headlines1(self):
        self.cross_compare(
            creole_string=r"""
                = Section Title 1

                == Section Title 2

                === Section Title 3

                ==== Section Title 4

                ===== Section Title 5

                ====== Section Title 6
            """,
            textile_string="""
                h1. Section Title 1

                h2. Section Title 2

                h3. Section Title 3

                h4. Section Title 4

                h5. Section Title 5

                h6. Section Title 6
            """,
            html_string="""
                <h1>Section Title 1</h1>

                <h2>Section Title 2</h2>

                <h3>Section Title 3</h3>

                <h4>Section Title 4</h4>

                <h5>Section Title 5</h5>

                <h6>Section Title 6</h6>
            """
        )
        self.cross_compare(
            rest_string="""
                ===============
                Section Title 1
                ===============

                ---------------
                Section Title 2
                ---------------

                Section Title 3
                ===============

                Section Title 4
                ---------------

                Section Title 5
                ```````````````

                Section Title 6
                '''''''''''''''
            """,
            html_string="""
                <h1>Section Title 1</h1>
                <h2>Section Title 2</h2>
                <h3>Section Title 3</h3>
                <h4>Section Title 4</h4>
                <h5>Section Title 5</h5>
                <h6>Section Title 6</h6>
            """
        )

    def test_horizontal_rule(self):
        all_markups = """
            Text before horizontal rule.

            ----

            Text after the line.
        """
        self.cross_compare(
            creole_string=all_markups,
            # textile_string=all_markups, # FIXME: textile and <hr> ?
            html_string="""
                <p>Text before horizontal rule.</p>

                <hr />

                <p>Text after the line.</p>
            """
        )
        self.cross_compare(
            rest_string=all_markups,
            html_string="""
                <p>Text before horizontal rule.</p>
                <hr />
                <p>Text after the line.</p>
            """
        )

    def test_link(self):
        self.cross_compare(
            creole_string=r"""
                X [[http://domain.tld|link B]] test.
            """,
            textile_string="""
                X "link B":http://domain.tld test.
            """,
            rest_string="""
                X `link B <http://domain.tld>`_ test.
            """,
            html_string="""
                <p>X <a href="http://domain.tld">link B</a> test.</p>
            """
        )

    def test_link_without_title(self):
        self.cross_compare(
            creole_string=r"""
                [[http://www.pylucid.org]]
            """,
            textile_string="""
                "http://www.pylucid.org":http://www.pylucid.org
            """,
            rest_string="""
                `http://www.pylucid.org <http://www.pylucid.org>`_
            """,
            html_string="""
                <p><a href="http://www.pylucid.org">http://www.pylucid.org</a></p>
            """
        )

    def test_link_with_unknown_protocol(self):
        self.cross_compare(
            creole_string=r"""
                X [[foo://bar|unknown protocol]] Y
            """,
            # textile will return '#' if url scheme is unknown!
            # textile_string="""
            #     X "unknown protocol":foo://bar Y
            # """,
            rest_string="""
                X `unknown protocol <foo://bar>`_ Y
            """,
            html_string="""
                <p>X <a href="foo://bar">unknown protocol</a> Y</p>
            """
        )

    def test_link_with_at_sign(self):
        self.cross_compare(
            creole_string=r"""
                X [[http://de.wikipedia.org/wiki/Creole_(Markup)|Creole@wikipedia]]
            """,
            html_string="""
                <p>X <a href="http://de.wikipedia.org/wiki/Creole_(Markup)">Creole@wikipedia</a></p>
            """
        )
        self.cross_compare(
            rest_string="""
                X `Creole@wikipedia <http://de.wikipedia.org/wiki/Creole_(Markup)>`_
            """,
            html_string="""
                <p>X <a href="http://de.wikipedia.org/wiki/Creole_(Markup)">Creole&#64;wikipedia</a></p>
            """
        )
        self.cross_compare_textile(
            textile_string="""
                X "foo@domain":http://domain.tld
            """,
            html_string="""
                <p>X <a href="http://domain.tld">foo@domain</a></p>
            """
        )

    def test_image(self):
        self.cross_compare(
            creole_string=r"""
                a {{/image.jpg|JPG pictures}} and
                a {{/image.jpeg|JPEG pictures}} and
                a {{/image.gif|GIF pictures}} and
                a {{/image.png|PNG pictures}} !
                {{/path1/path2/image|Image without files ext?}}
            """,
            html_string="""
                <p>a <img src="/image.jpg" title="JPG pictures" alt="JPG pictures" /> and<br />
                a <img src="/image.jpeg" title="JPEG pictures" alt="JPEG pictures" /> and<br />
                a <img src="/image.gif" title="GIF pictures" alt="GIF pictures" /> and<br />
                a <img src="/image.png" title="PNG pictures" alt="PNG pictures" /> !<br />
                <img src="/path1/path2/image" title="Image without files ext?" alt="Image without files ext?" /></p>
            """
        )
        self.cross_compare(
            textile_string="""
                a !/image.jpg(JPG pictures)! and
                a !/image.jpeg(JPEG pictures)! and
                a !/image.gif(GIF pictures)! and
                a !/image.png(PNG pictures)! !
                !/path1/path2/image(Image without files ext?)!
            """,
            html_string="""
                <p>a <img alt="JPG pictures" src="/image.jpg" title="JPG pictures" /> and<br />

                a <img alt="JPEG pictures" src="/image.jpeg" title="JPEG pictures" /> and<br />

                a <img alt="GIF pictures" src="/image.gif" title="GIF pictures" /> and<br />

                a <img alt="PNG pictures" src="/image.png" title="PNG pictures" /> !<br />

                <img alt="Image without files ext?" src="/path1/path2/image" title="Image without files ext?" /></p>
            """
        )
        self.cross_compare(
            rest_string="""
                1 |JPG pictures| one

                .. |JPG pictures| image:: /image.jpg

                2 |JPEG pictures| two

                .. |JPEG pictures| image:: /image.jpeg

                3 |GIF pictures| tree

                .. |GIF pictures| image:: /image.gif

                4 |PNG pictures| four

                .. |PNG pictures| image:: /image.png

                5 |Image without files ext?| five

                .. |Image without files ext?| image:: /path1/path2/image
            """,
            html_string="""
                <p>1 <img alt="JPG pictures" src="/image.jpg" /> one</p>
                <p>2 <img alt="JPEG pictures" src="/image.jpeg" /> two</p>
                <p>3 <img alt="GIF pictures" src="/image.gif" /> tree</p>
                <p>4 <img alt="PNG pictures" src="/image.png" /> four</p>
                <p>5 <img alt="Image without files ext?" src="/path1/path2/image" /> five</p>
            """
        )

    def test_link_image(self):
        """ FIXME: ReSt. and linked images """
        self.cross_compare(
            creole_string=r"""
                Linked [[http://example.com/|{{myimage.jpg|example site}} image]]
            """,
            html_string="""
                <p>Linked <a href="http://example.com/"><img src="myimage.jpg" title="example site" alt="example site" /> image</a></p>
            """
        )
        self.cross_compare(
            textile_string="""
                Linked "!myimage.jpg(example site)! image":http://example.com/
            """,
            html_string="""
                <p>Linked <a href="http://example.com/"><img alt="example site" src="myimage.jpg" title="example site" /> image</a></p>
            """
        )


#        self.cross_compare(# FIXME: ReSt
#            rest_string="""
#                I recommend you try |PyLucid CMS|_.
#
#                .. |PyLucid CMS| image:: /images/pylucid.png
#                .. _PyLucid CMS: http://www.pylucid.org/
#            """,
#            html_string="""
#                <p>I recommend you try <a href="http://www.pylucid.org/"><img alt="PyLucid CMS" src="/images/pylucid.png" /></a>.</p>
#            """
#        )

    def test_pre1(self):
        self.cross_compare(
            creole_string=r"""
                {{{
                * no list
                }}}
                """,
            textile_string="""
                <pre>
                * no list
                </pre>
                """,
            html_string="""
                <pre>
                * no list
                </pre>
            """)
        self.cross_compare(  # FIXME: Not the best html2rest output
            rest_string="""
                Preformatting text:

                ::

                    Here some performatting with
                    no `link <http://domain.tld>`_ here.
                    text... end.

                Under pre block
            """,
            html_string="""
                <p>Preformatting text:</p>
                <pre>
                Here some performatting with
                no `link &lt;http://domain.tld&gt;`_ here.
                text... end.
                </pre>
                <p>Under pre block</p>
            """
        )


#    def test_pre2(self):
#        """ TODO: html2creole: wrong lineendings """
#        self.cross_compare(
#            creole_string=r"""
#                start
#
#                {{{
#                * no list
#                }}}
#
#                end
#                """,
#            textile_string="""
#                start
#
#                <pre>
#                * no list
#                </pre>
#
#                end
#                """,
#            html_string="""
#                <p>start</p>
#
#                <pre>
#                * no list
#                </pre>
#
#                <p>end</p>
#            """)

    def test_pre_contains_braces(self):
        self.cross_compare(
            creole_string="""
                {{{
                # Closing braces in nowiki:
                if (x != NULL) {
                  for (i = 0) {
                    if (x = 1) {
                      x[i]--;
                  }}}
                }}}
                """,
            textile_string="""
                <pre>
                # Closing braces in nowiki:
                if (x != NULL) {
                  for (i = 0) {
                    if (x = 1) {
                      x[i]--;
                  }}}
                </pre>
                """,
            rest_string="""
                ::

                    # Closing braces in nowiki:
                    if (x != NULL) {
                      for (i = 0) {
                        if (x = 1) {
                          x[i]--;
                      }}}
                """,
            html_string="""
                <pre>
                # Closing braces in nowiki:
                if (x != NULL) {
                  for (i = 0) {
                    if (x = 1) {
                      x[i]--;
                  }}}
                </pre>
            """)

    def test_list(self):
        """ Bold, Italics, Links, Pre in Lists """
        self.cross_compare(
            creole_string=r"""
                * **bold** item
                * //italic// item

                # item about a [[/foo/bar|certain_page]]
                """,
            textile_string="""
                * *bold* item
                * __italic__ item

                # item about a "certain_page":/foo/bar
            """,
            html_string="""
                <ul>
                  <li><strong>bold</strong> item</li>
                  <li><i>italic</i> item</li>
                </ul>
                <ol>
                  <li>item about a <a href="/foo/bar">certain_page</a></li>
                </ol>
            """,
            strip_lines=True
        )

    def test_simple_table(self):
        self.cross_compare(
            creole_string="""
                |= Headline 1 |= Headline 2 |
                | cell one    | cell two    |
                """,
            textile_string="""
                |_. Headline 1|_. Headline 2|
                |cell one|cell two|
                """,
            html_string="""
                <table>
                <tr>
                    <th>Headline 1</th>
                    <th>Headline 2</th>
                </tr>
                <tr>
                    <td>cell one</td>
                    <td>cell two</td>
                </tr>
                </table>
            """,
            # debug=True
            strip_lines=True,
        )
        self.cross_compare(
            rest_string="""
                +------------+------------+
                | Headline 1 | Headline 2 |
                +============+============+
                | cell one   | cell two   |
                +------------+------------+
                """,
            html_string="""
                <table>
                <colgroup>
                <col width="50%" />
                <col width="50%" />
                </colgroup>
                <tr><th>Headline 1</th>
                <th>Headline 2</th>
                </tr>
                <tr><td>cell one</td>
                <td>cell two</td>
                </tr>
                </table>
            """,
            # debug=True
        )


if __name__ == '__main__':
    unittest.main()



"""
    Unittest which failed, cause bugfixes not implemented, yet.
"""

import unittest

from creole.html_tools.strip_html import strip_html
from creole.tests.utils.base_unittest import BaseCreoleTest


class StripHtml(unittest.TestCase):
    @unittest.expectedFailure
    def test_not_closed_image_tag(self):
        output = strip_html('<p>a <img src="/image.jpg"> image.</p>')
        self.assertEqual(output, '<p>a <img src="/image.jpg"> image.</p>')

    @unittest.expectedFailure
    def test_remove_linebreak(self):
        output = strip_html('<strong>foo</strong>\n<ul><li>one</li></ul>')
        self.assertEqual(output, '<strong>foo</strong><ul><li>one</li></ul>')


class CrossCompareCreoleTests(BaseCreoleTest):
    @unittest.expectedFailure
    def test_cross_lines_creole2html(self):
        """ TODO: bold/italics cross lines in creole2html
        see: http://code.google.com/p/python-creole/issues/detail?id=13
        Info: The way html2creole works, see above
        """
        self.cross_compare_creole(
            creole_string=r"""
                Bold and italics should //be
                able// to **cross
                lines.**
            """,
            html_string="""
                <p>Bold and italics should <i>be<br />
                able</i> to <strong>cross<br />
                lines.</strong></p>
            """
        )

    @unittest.expectedFailure
    def test_cross_paragraphs(self):
        """ TODO: bold/italics cross paragraphs in creole2html
        see: http://code.google.com/p/python-creole/issues/detail?id=13
        """
        self.assert_creole2html("""
            But, should //not be...

            ...able// to cross paragraphs.
        """, """
            <p>But, should <em>not be...</em></p>
            <p>...able<em> to cross paragraphs.</em></p>
        """)

    @unittest.expectedFailure
    def test_escape_inline(self):
        """ TODO: different pre/code syntax?
        """
        self.cross_compare_creole(r"""
            this is {{{**escaped** inline}}}, isn't it?

            {{{
            a **code**
            block
            }}}
        """, """
            <p>this is <tt>**escaped** inline</tt>, isn't it?</p>

            <pre>
            a **code**
            block
            </pre>
        """)


class TestHtml2CreoleMarkup(BaseCreoleTest):
    @unittest.expectedFailure
    def test_format_in_a_text(self):
        """ TODO: http://code.google.com/p/python-creole/issues/detail?id=4 """
        self.assert_html2creole(r"""
            **[[/url/|title]]**
        """, """
            <a href="/url/"><strong>title</strong></a>
        """)

    @unittest.expectedFailure
    def test_newline_before_headline(self):
        """ TODO: http://code.google.com/p/python-creole/issues/detail?id=16#c5 """
        self.assert_html2creole(r"""
            **foo**

            = one
        """, """
            <b>foo</b>
            <h1>one</h1>
        """)  # , debug=True)

    @unittest.expectedFailure
    def test_no_space_before_blocktag(self):
        """ TODO: Bug in html2creole.strip_html(): Don't add a space before/after block tags """
        self.assert_html2creole(r"""
            **foo**

            * one
        """, """
            <b>foo</b>
            <ul><li>one</li></ul>
        """  # , debug=True
                                )

    @unittest.expectedFailure
    def test_escape_char(self):
        self.assert_html2creole(r"""
            ~#1
            http://domain.tld/~bar/
            ~http://domain.tld/
            [[Link]]
            ~[[Link]]
        """, """
            <p>#1<br />
            <a href="http://domain.tld/~bar/">http://domain.tld/~bar/</a><br />
            http://domain.tld/<br />
            <a href="Link">Link</a><br />
            [[Link]]</p>
        """)

    @unittest.expectedFailure
    def test_images(self):
        self.assert_html2creole(r"""
            a {{/image.jpg|JPG pictures}} and
            a {{/image.jpeg|JPEG pictures}} and
            a {{/image.gif|GIF pictures}} and
            a {{/image.png|PNG pictures}} !

            picture [[www.domain.tld|{{foo.JPG|Foo}}]] as a link
        """, """
            <p>a <img src="/image.jpg" alt="JPG pictures"> and<br />
            a <img src="/image.jpeg" alt="JPEG pictures"> and<br />
            a <img src="/image.gif" alt="GIF pictures" /> and<br />
            a <img src="/image.png" alt="PNG pictures" /> !</p>

            <p>picture <a href="www.domain.tld"><img src="foo.JPG" alt="Foo"></a> as a link</p>
        """  # , debug=True
                                )



"""
    python-creole utils
    ~~~~~~~~~~~~~~~~~~~


    :copyleft: 2008-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import re
from html import entities


entities_rules = '|'.join([
    r"(&\#(?P<number>\d+);)",
    r"(&\#x(?P<hex>[a-fA-F0-9]+);)",
    r"(&(?P<named>[a-zA-Z]+);)",
])
# print(entities_rules)
entities_regex = re.compile(
    entities_rules, re.VERBOSE | re.UNICODE | re.MULTILINE
)


class Deentity:
    """
    replace html entity

    >>> d = Deentity()
    >>> d.replace_all("-=[&nbsp;&gt;&#62;&#x3E;nice&lt;&#60;&#x3C;&nbsp;]=-")
    '-=[ >>>nice<<< ]=-'

    >>> d.replace_all("-=[M&uuml;hlheim]=-") # uuml - latin small letter u with diaeresis
    '-=[M\\xfchlheim]=-'

    >>> d.replace_number("126")
    '~'
    >>> d.replace_hex("7E")
    '~'
    >>> d.replace_named("amp")
    '&'
    """

    def replace_number(self, text):
        """ unicode number entity """
        unicode_no = int(text)
        return chr(unicode_no)

    def replace_hex(self, text):
        """ hex entity """
        unicode_no = int(text, 16)
        return chr(unicode_no)

    def replace_named(self, text):
        """ named entity """
        if text == "nbsp":
            # Non breaking spaces is not in htmlentitydefs
            return " "
        else:
            codepoint = entities.name2codepoint[text]
            return chr(codepoint)

    def replace_all(self, content):
        """ replace all html entities form the given text. """
        def replace_entity(match):
            groups = match.groupdict()
            for name, text in groups.items():
                if text is not None:
                    replace_method = getattr(self, f'replace_{name}')
                    return replace_method(text)

            # Should never happen:
            raise RuntimeError("deentitfy re rules wrong!")

        return entities_regex.sub(replace_entity, content)


if __name__ == '__main__':
    import doctest
    print(doctest.testmod())



"""
    unittest
    ~~~~~~~~

    :copyleft: 2015-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import sys
import unittest

from creole.tests.utils.unittest_subprocess import SubprocessMixin


class TestSubprocessMixin(unittest.TestCase, SubprocessMixin):

    def test_subprocess(self):
        popen_args, retcode, stdout = self.subprocess(
            popen_args=[sys.executable, "-c", "import sys;sys.stdout.write('to stdout')"],
        )
        self.assertEqual(stdout, "to stdout")
        self.assertEqual(retcode, 0)

    def test_assertSubprocess(self):
        code = (
            "import sys;"
            "sys.stdout.write('to stdout 1\\n');"
            "sys.stdout.flush();"
            "sys.stderr.write('to stderr 1\\n');"
            "sys.stderr.flush();"
            "sys.stdout.write('to stdout 2\\n');"
            "sys.stdout.flush();"
            "sys.stderr.write('to stderr 2\\n');"
            "sys.stderr.flush();"
        )
        self.assertSubprocess(
            popen_args=(sys.executable, "-c", code),
            retcode=0,
            stdout=(
                "to stdout 1\n"
                "to stderr 1\n"
                "to stdout 2\n"
                "to stderr 2\n"
            ),
        )



"""
    Python setup.py utilities
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Generate ReStructuredText README from README.creole.
    Usable for other python packages, too.

    More information:
        https://github.com/jedie/python-creole/wiki/Use-In-Setup

    :copyleft: 2011-2020 by the python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import codecs
import datetime
import os
import sys
import warnings
from pathlib import Path

from readme_renderer.rst import render

from creole import creole2html, html2rest
from creole.shared.diff_utils import unified_diff
from creole.shared.unknown_tags import raise_unknown_node, transparent_unknown_nodes


RAISE_ERRORS_ARGS = (
    'check', 'register', 'sdist', 'bdist', 'upload',
    '--long-description', '--restructuredtext',
)


def should_raise_errors():
    """
    Raise only errors, if one of RAISE_ERRORS_ARGS is in sys.argv
    or if no argument presents.
    """
    if len(sys.argv) == 1:
        return True

    for arg in RAISE_ERRORS_ARGS:
        if arg in sys.argv:
            return True
    return False


def get_long_description(package_root, filename='README.creole', raise_errors=None):
    """ read README file and convert it on-the-fly to ReStructuredText. """

    warnings.warn('get_long_description() will be removed in the future', DeprecationWarning)

    if raise_errors is None:
        raise_errors = should_raise_errors()

    if raise_errors:
        sys.stderr.write('Test creole2rest and raise an error, if rendering failed...\n')
        # raise a error if a unknown node found
        unknown_emit = raise_unknown_node
    else:
        # ignore unknown nodes
        unknown_emit = transparent_unknown_nodes

    filepath = os.path.join(package_root, filename)
    long_description_origin = ''
    try:
        # Read creole README
        f = codecs.open(filepath, 'r', encoding='utf-8')
        long_description_origin = f.read().strip()
        f.close()

        # convert creole into html
        long_description_html = creole2html(long_description_origin)

        # convert html to ReSt
        long_description_rest = html2rest(
            long_description_html,
            emitter_kwargs={'unknown_emit': unknown_emit}
        )
    except Exception:
        if raise_errors:
            raise
        # Don't raise the error e.g. in ./setup install process
        evalue = sys.exc_info()[1]
        long_description_rest = f'[Error: {evalue}]\n{long_description_origin}'
    else:
        if raise_errors:
            # Test created ReSt code like PyPi does it.
            from creole.rest_tools.pypi_rest2html import pypi_rest2html
            try:
                pypi_rest2html(long_description_rest)
            except SystemExit as e:
                msg = f'Error creole2rest self test failed: rest2html() exist with status code: {e.args[0]}\n'
                sys.stderr.write(msg)
                sys.exit(msg)
            except Exception as e:
                sys.exit(f'ReSt2html error: {e}')
            else:
                if 'check' in sys.argv:
                    print('Generating creole to ReSt to html, ok.')

    return long_description_rest


def _generate_rst_readme(*, creole_readme_path):
    with creole_readme_path.open('r') as f:
        creole_readme = f.read().strip()

    # convert creole into html
    html_readme = creole2html(creole_readme)

    # convert html to ReSt
    rest_readme = html2rest(
        html_readme,
        emitter_kwargs={
            'unknown_emit': raise_unknown_node  # raise a error if a unknown node found
        }
    )

    # Check if generated ReSt is valid, see also:
    # https://pypi.org/help/#description-content-type
    rendered = render(rest_readme, stream=sys.stderr)
    if rendered is None:
        sys.exit(1)

    return rest_readme


def update_rst_readme(package_root, filename='README.creole'):
    """
    Generate README.rst from README.creole
    """
    assert isinstance(package_root, Path)
    assert package_root.is_dir(), f'Directory not found: {package_root}'
    creole_readme_path = Path(package_root, filename)
    assert creole_readme_path.is_file(), f'File not found: {creole_readme_path}'

    rest_readme_path = creole_readme_path.with_suffix('.rst')
    print(
        f'Generate {rest_readme_path.name} from {creole_readme_path.name}',
        end='...', flush=True
    )

    rest_readme = _generate_rst_readme(creole_readme_path=creole_readme_path)

    # Check if content was changed
    changed = False
    with rest_readme_path.open('r') as f:
        for new_line, old_line in zip(rest_readme.splitlines(), f):
            if new_line.rstrip() != old_line.rstrip():
                changed = True
                break

    if not changed:
        # The existing README.rst is up-to-date: Don't change the timestamp
        print('nothing changed, ok.')
        return rest_readme_path

    with rest_readme_path.open('w') as f:
        f.write(rest_readme)

        # Add a note about generation with modification time from source:

        f.write('\n\n------------\n\n')

        modification_time = creole_readme_path.stat().st_mtime
        dt = datetime.datetime.fromtimestamp(modification_time)
        dt = dt.replace(microsecond=0)
        dt = dt.isoformat(sep=' ')
        f.write(f'``Note: this file is generated from {filename} {dt} with "python-creole"``')

    print('done.')
    return rest_readme_path


def assert_rst_readme(package_root, filename='README.creole'):
    """
    raise AssertionError if README.rst is not up-to-date.
    """
    creole_readme_path = Path(package_root, filename)
    rest_readme = _generate_rst_readme(creole_readme_path=creole_readme_path)
    rest_readme_path = creole_readme_path.with_suffix('.rst')
    with rest_readme_path.open('r') as f:
        content = f.read()

    assert len(content) > 0, f'Empty content in {rest_readme_path}'
    content = content.rsplit('\n', 4)[0]  # remove note about generation with modification time

    if rest_readme != content:
        diff = unified_diff(content, rest_readme, filename=rest_readme_path.name)
        raise AssertionError(f'{rest_readme_path.name} is not up-to-date:\n{diff}')


def update_creole_rst_readme():
    return update_rst_readme(
        package_root=Path(__file__).parent.parent,
        filename='README.creole'
    )


if __name__ == '__main__':
    update_creole_rst_readme()



"""
    html -> textile Emitter
    ~~~~~~~~~~~~~~~~~~~~~~


    :copyleft: 2008-2011 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import posixpath

from creole.shared.base_emitter import BaseEmitter


class TextileEmitter(BaseEmitter):
    """
    Build from a document_tree (html2creole.parser.HtmlParser instance) a
    creole markup text.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.table_head_prefix = "_. "
        self.table_auto_width = False

    def emit(self):
        """Emit the document represented by self.root DOM tree."""
        return self.emit_node(self.root).strip()  # FIXME

    # --------------------------------------------------------------------------

    def blockdata_pre_emit(self, node):
        """ pre block -> with newline at the end """
        return f"<pre>{self.deentity.replace_all(node.content)}</pre>\n"

    def inlinedata_pre_emit(self, node):
        """ a pre inline block -> no newline at the end """
        return f"<pre>{self.deentity.replace_all(node.content)}</pre>"

    def blockdata_pass_emit(self, node):
        return f"{node.content}\n\n"

    # --------------------------------------------------------------------------

    def p_emit(self, node):
        return f"{self.emit_children(node)}\n\n"

    def headline_emit(self, node):
        return f"h{node.level:d}. {self.emit_children(node)}\n\n"

    # --------------------------------------------------------------------------

    def _typeface(self, node, key):
        return key + self.emit_children(node) + key

    def strong_emit(self, node):
        return self._typeface(node, key="*")

    def b_emit(self, node):
        return self._typeface(node, key="**")
    big_emit = strong_emit

    def i_emit(self, node):
        return self._typeface(node, key="__")

    def em_emit(self, node):
        return self._typeface(node, key="_")

    def sup_emit(self, node):
        return self._typeface(node, key="^")

    def sub_emit(self, node):
        return self._typeface(node, key="~")

    def del_emit(self, node):
        return self._typeface(node, key="-")

    def cite_emit(self, node):
        return self._typeface(node, key="??")

    def ins_emit(self, node):
        return self._typeface(node, key="+")

    def span_emit(self, node):
        return self._typeface(node, key="%")

    def code_emit(self, node):
        return self._typeface(node, key="@")

    # --------------------------------------------------------------------------

    def hr_emit(self, node):
        return "----\n\n"

    def a_emit(self, node):
        link_text = self.emit_children(node)
        url = node.attrs["href"]
        return f'"{link_text}":{url}'

    def img_emit(self, node):
        src = node.attrs["src"]

        if src.split(':')[0] == 'data':
            return ""

        title = node.attrs.get("title", "")
        alt = node.attrs.get("alt", "")
        if len(alt) > len(title):  # Use the longest one
            text = alt
        else:
            text = title

        if text == "":  # Use filename as picture text
            text = posixpath.basename(src)

        return f"!{src}({text})!"

    # --------------------------------------------------------------------------

    def ul_emit(self, node):
        return self._list_emit(node, list_type="*")

    def ol_emit(self, node):
        return self._list_emit(node, list_type="#")



"""
    unittest for some utils
    ~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2011 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import unittest

from creole.shared.markup_table import MarkupTable
from creole.tests.utils.utils import MarkupTest


class UtilsTests(MarkupTest):

    def test_assertEqual(self):
        self.assertRaises(
            AssertionError, self.assertEqual, "foo", "bar"
        )

    def test_prepare_text_base(self):
        out1 = self._prepare_text("""
            one line
            line two""")
        self.assertEqual(out1, "one line\nline two")

        out2 = self._prepare_text("""
            one line
            line two
        """)
        self.assertEqual(out2, "one line\nline two")

    def test_prepare_text_last_line_empty(self):

        out3 = self._prepare_text("""
            one line
            line two

        """)
        self.assertEqual(out3, "one line\nline two\n")

    def test_prepare_text_empty_line(self):
        self.assertEqual(self._prepare_text("""
            line one

            line two
        """), "line one\n\nline two")

    def test_prepare_text_first_line_empty(self):
        self.assertEqual(self._prepare_text("""

            line one
            line two
        """), "\nline one\nline two")

    def test_prepare_space_and_line_end(self):
        self.assertEqual(self._prepare_text("\n  111  \n  222"), "111\n222")

    def test_prepare_text_indent_line(self):
        self.assertEqual("line one\n  indent line\nline two", self._prepare_text("""
            line one
              indent line
            line two
        """))

    def test_prepare_text_indent_with_empty_end(self):
        out4 = self._prepare_text("""
            one line
                line two

        """)
        self.assertEqual(out4, "one line\n    line two\n")

    def assertEqual2(self, first, second, msg=""):
        self.assertNotEqual(first, second, msg)

#        first = self._prepare_text(first)
        second = self._prepare_text(second)

        self.assertEqual(first, second, msg)

    def test_markup_table_creole(self):
        t = MarkupTable(head_prefix="* ")
        t.add_tr()
        t.add_th("head1")
        t.add_th("head2")
        t.add_tr()
        t.add_td("1.1.")
        t.add_td("1.2.")
        t.add_tr()
        t.add_td("2.1.")
        t.add_td("2.2.")
        table = t.get_table_markup()

        self.assertEqual2(
            table,
            """
            |* head1 |* head2 |
            | 1.1.   | 1.2.   |
            | 2.1.   | 2.2.   |
            """
        )

    def test_markup_table_textile(self):
        t = MarkupTable(head_prefix="_. ", auto_width=False)
        t.add_tr()
        t.add_th("head1")
        t.add_th("head2")
        t.add_tr()
        t.add_td("1.1.")
        t.add_td("1.2.")
        t.add_tr()
        t.add_td("2.1.")
        t.add_td("2.2.")
        table = t.get_table_markup()

        self.assertEqual2(
            table,
            """
            |_. head1|_. head2|
            |1.1.|1.2.|
            |2.1.|2.2.|
            """
        )

    def test_markup_table_rest(self):
        t = MarkupTable(head_prefix="")
        t.add_tr()
        t.add_th("head1")
        t.add_th("head2")
        t.add_tr()
        t.add_td("1.1.")
        t.add_td("1.2.")
        t.add_tr()
        t.add_td("2.1.")
        t.add_td("2.2.")
        table = t.get_rest_table()

        self.assertEqual2(
            table,
            """
            +-------+-------+
            | head1 | head2 |
            +=======+=======+
            | 1.1.  | 1.2.  |
            +-------+-------+
            | 2.1.  | 2.2.  |
            +-------+-------+
            """
        )


if __name__ == '__main__':
    unittest.main()



"""
    unitest base class
    ~~~~~~~~~~~~~~~~~~

    Basic unittest class for all python-creole tests.

    :copyleft: 2008-2020 by python-creole team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import re

import textile

from creole import creole2html, html2creole, html2rest, html2textile
from creole.rest_tools.clean_writer import rest2html
from creole.tests.utils.utils import MarkupTest


tabs2spaces_re = re.compile(r"^(\t*)(.*?)$", re.M)


def tabs2spaces(html):
    """ form reformating textile html code
    >>> tabs2spaces("\\t<p>one<br />\\n\\t\\ttwo<br />\\n\\t\\t\\ttree</p>")
    '<p>one<br />\\n  two<br />\\n    tree</p>'
    """
    def reformat_tabs(match):
        tabs = match.group(1)
        text = match.group(2)

        indent = len(tabs) - 1
        if indent < 0:
            indent = 0

#        print(len(tabs), indent, repr(tabs), text)
        return "  " * indent + text
    return tabs2spaces_re.sub(reformat_tabs, html)


def strip_html_lines(html, strip_lines=False):
    """
    >>> strip_html_lines("\t<p>foo   \\n\\n\t\t  bar</p>", strip_lines=True)
    '<p>foo\\nbar</p>'
    """
    html = "\n".join(
        [line.strip(" \t") for line in html.splitlines() if line]
    )
    return html


class BaseCreoleTest(MarkupTest):
    """
    Basic unittest class for all python-creole unittest classes.
    """

    def _debug_text(self, msg, raw_text):
        text = raw_text.replace(" ", ".")
        text = text.replace("\n", "\\n\n")
        text = text.replace("\t", "\\t")

        print()
        print("_" * 79)
        print(f" Debug Text: {msg}:")
        print(text)
        print("-" * 79)

    def assert_creole2html(
        self, raw_creole, raw_html,
        strip_lines=False, debug=True,
        parser_kwargs={}, emitter_kwargs={},
        block_rules=None, blog_line_breaks=True, macros=None, verbose=True, stderr=None,
        strict=False,
    ):
        """
        compare the generated html code from the markup string >creole_string<
        with the >html_string< reference.
        """
        self.assertNotEqual(raw_creole, raw_html)
        self.assertEqual(parser_kwargs, {}, "parser_kwargs is deprecated!")
        self.assertEqual(emitter_kwargs, {}, "parser_kwargs is deprecated!")

        # prepare whitespace on test strings
        markup_string = self._prepare_text(raw_creole)
        assert isinstance(markup_string, str)

        html_string = self._prepare_text(raw_html)
        assert isinstance(html_string, str)
        if strip_lines:
            html_string = strip_html_lines(html_string, strip_lines)
        self._debug_text("assert_creole2html() html_string reference", html_string)

        # convert creole markup into html code
        out_string = creole2html(
            markup_string, debug=debug,
            block_rules=block_rules, blog_line_breaks=blog_line_breaks,
            macros=macros, verbose=verbose, stderr=stderr,
            strict=strict,
        )
        self._debug_text("assert_creole2html() creole2html output", out_string)

        if strip_lines:
            out_string = strip_html_lines(out_string, strip_lines)
        else:
            out_string = out_string.replace("\t", "    ")

        # compare
        self.assertEqual(out_string, html_string, msg="creole2html")

    def assert_html2creole2(self, creole, html,
                            debug=True,
                            unknown_emit=None,
                            strict=False,
                            ):
        # convert html code into creole markup
        out_string = html2creole(
            html, debug, unknown_emit=unknown_emit, strict=strict
        )
        self._debug_text("assert_html2creole() html2creole", out_string)

        # compare
        self.assertEqual(out_string, creole, msg="html2creole")

    def assert_html2creole(self, raw_creole, raw_html,
                           strip_lines=False, debug=True,
                           # OLD API:
                           parser_kwargs={}, emitter_kwargs={},
                           # html2creole:
                           unknown_emit=None,
                           strict=False,
                           ):
        """
        Compare the genereted markup from the given >raw_html< html code, with
        the given >creole_string< reference string.
        """
        self.assertEqual(parser_kwargs, {}, "parser_kwargs is deprecated!")
        self.assertEqual(emitter_kwargs, {}, "parser_kwargs is deprecated!")
#        assert isinstance(raw_html, str)
#        creole_string = unicode(creole_string, encoding="utf8")
#        raw_html = unicode(raw_html, "utf8")

        self.assertNotEqual(raw_creole, raw_html)

        # prepare whitespace on test strings
        creole = self._prepare_text(raw_creole)
        assert isinstance(creole, str)
        if debug:
            self._debug_text("assert_creole2html() markup", creole)

        html = self._prepare_text(raw_html)
        assert isinstance(html, str)

        self.assert_html2creole2(creole, html, debug, unknown_emit, strict)

    def cross_compare_creole(self, creole_string, html_string,
                             strip_lines=False, debug=True,
                             # creole2html old API:
                             creole_parser_kwargs={}, html_emitter_kwargs={},
                             # html2creole old API:
                             html_parser_kwargs={}, creole_emitter_kwargs={},

                             # creole2html new API:
                             block_rules=None, blog_line_breaks=True, macros=None, stderr=None,
                             # html2creole:
                             unknown_emit=None
                             ):
        """
        Cross compare with:
            * creole2html
            * html2creole
        """
        self.assertEqual(creole_parser_kwargs, {}, "creole_parser_kwargs is deprecated!")
        self.assertEqual(html_emitter_kwargs, {}, "html_emitter_kwargs is deprecated!")
        self.assertEqual(html_parser_kwargs, {}, "html_parser_kwargs is deprecated!")
        self.assertEqual(creole_emitter_kwargs, {}, "creole_emitter_kwargs is deprecated!")

        assert isinstance(creole_string, str)
        assert isinstance(html_string, str)
        self.assertNotEqual(creole_string, html_string)

        self.assert_creole2html(
            raw_creole=creole_string, raw_html=html_string,
            strip_lines=strip_lines,
            block_rules=block_rules, blog_line_breaks=blog_line_breaks,
            macros=macros, stderr=stderr,
        )

        self.assert_html2creole(
            raw_creole=creole_string, raw_html=html_string, strip_lines=strip_lines,
            unknown_emit=unknown_emit,
        )

    def assert_html2textile(self, textile_string, html_string,
                            strip_lines=False, debug=False, parser_kwargs={}, emitter_kwargs={}):
        """
        Check html2textile
        """
        self.assertNotEqual(textile_string, html_string)

        textile_string = self._prepare_text(textile_string)
        html_string = self._prepare_text(html_string)

        if strip_lines:
            html_string = strip_html_lines(html_string, strip_lines)

        # compare html -> textile
        textile_string2 = html2textile(html_string, debug, parser_kwargs, emitter_kwargs)
        if debug:
            print("-" * 79)
            print(textile_string2)
            print("-" * 79)

        self.assertEqual(textile_string2, textile_string, msg="html2textile")

        return textile_string, html_string

    def cross_compare_textile(self, textile_string, html_string,
                              strip_lines=False, debug=False, parser_kwargs={}, emitter_kwargs={}):
        """
            Checks:
                * html2textile
                * textile2html
        """
#        assert isinstance(textile_string, str)
#        assert isinstance(html_string, str)
        self.assertNotEqual(textile_string, html_string)

        # compare html -> textile
        textile_string, html_string = self.assert_html2textile(
            textile_string, html_string,
            strip_lines, debug, parser_kwargs, emitter_kwargs
        )

        # compare textile -> html
        html = textile.textile(textile_string)
        html = html.replace("<br />", "<br />\n")
        html = tabs2spaces(html)
        if strip_lines:
            html = strip_html_lines(html, strip_lines)

        self.assertEqual(html_string, html, msg="textile2html")

    def assert_html2rest(self, rest_string, html_string,
                         strip_lines=False, debug=False, parser_kwargs={}, emitter_kwargs={}):
        """
        Check html to reStructuredText converter
        """
        self.assertNotEqual(rest_string, html_string)

        rest_string = self._prepare_text(rest_string)
        print("-" * 100)
        print(rest_string)
        html_string = self._prepare_text(html_string)
        if strip_lines:
            html_string = strip_html_lines(html_string, strip_lines)
        print("-" * 100)
        print(html_string)

        # compare html -> reStructuredText
        rest_string2 = html2rest(html_string, debug, parser_kwargs, emitter_kwargs)
        print("-" * 100)
        print(rest_string2)
        print("-" * 100)

        self.assertEqual(rest_string2, rest_string, msg="html2rest")

        return rest_string, html_string

    def assert_rest2html(self, rest_string, html_string,
                         strip_lines=False, debug=False, prepare_strings=True, **kwargs):

        # compare rest -> html
        if prepare_strings:
            rest_string = self._prepare_text(rest_string)
            html_string = self._prepare_text(html_string)

        html = rest2html(rest_string, **kwargs)

        if debug:
            print(rest_string)
            print(html_string)
            print(html)

        html = html.strip()
#        html = html.replace("<br />", "<br />\n")
#        html = tabs2spaces(html)
        if strip_lines:
            html = strip_html_lines(html, strip_lines)

        self.assertEqual(html, html_string, msg="rest2html")

    def cross_compare_rest(self, rest_string, html_string,
                           strip_lines=False, debug=False, parser_kwargs={}, emitter_kwargs={}):
        #        assert isinstance(textile_string, str)
        #        assert isinstance(html_string, str)
        self.assertNotEqual(rest_string, html_string)

        rest_string, html_string = self.assert_html2rest(
            rest_string, html_string,
            strip_lines, debug, parser_kwargs, emitter_kwargs
        )

        # compare rest -> html
        self.assert_rest2html(
            rest_string, html_string,
            strip_lines=strip_lines, debug=debug,
            prepare_strings=False,
        )

    def cross_compare(self,
                      html_string,
                      creole_string=None,
                      textile_string=None,
                      rest_string=None,
                      strip_lines=False, debug=False, parser_kwargs={}, emitter_kwargs={}):
        """
        Cross compare with:
            * creole2html
            * html2creole
            * html2textile
            * html2ReSt
        """
        if creole_string:
            self.cross_compare_creole(
                creole_string, html_string, strip_lines, debug, parser_kwargs, emitter_kwargs
            )

        if textile_string:
            self.cross_compare_textile(
                textile_string, html_string, strip_lines, debug, parser_kwargs, emitter_kwargs
            )

        if rest_string:
            self.cross_compare_rest(
                rest_string, html_string, strip_lines, debug, parser_kwargs, emitter_kwargs
            )


if __name__ == '__main__':
    import doctest
    print(doctest.testmod())
